# -*- coding: utf-8 -*-
""" EMOTIC_VAD is the base class for instantiating various end-to-end models for continuous emotion recognition in VAD space
    using the EMOTIC dataset.

# Reference
- [Emotion Recognition in Context](http://sunai.uoc.edu/emotic/pdf/EMOTIC_cvpr2017.pdf)
- https://stackoverflow.com/questions/43452441/keras-all-layer-names-should-be-unique

"""


from __future__ import print_function
import numpy as np
import time

import h5py

from utils.generic_utils import hms_string
from functools import partial

from keras.engine.topology import get_source_inputs
from keras.models import Model
from keras.optimizers import SGD, RMSprop
from keras.applications.imagenet_utils import _obtain_input_shape
from keras.layers import Dense, GlobalAveragePooling2D, Dropout, BatchNormalization, Activation
from keras.layers.merge import concatenate
from keras.callbacks import ModelCheckpoint, EarlyStopping, CSVLogger
from keras.layers import Input
from keras import regularizers
from keras.applications.vgg16 import VGG16
from keras.applications.vgg19 import VGG19
from keras.applications.resnet50 import ResNet50
from applications.vgg16_places_365 import VGG16_Places365

from keras.losses import binary_crossentropy

from preprocessing.emotic.custom_generator import custom_generator, custom_generator_single_output
from utils.generic_utils import rmse, euclidean_distance_loss


class EMOTIC_VAD():
    """Loads the parameters needed for the training process on class instantiation
        & sets out the training process of various models (defined as different functions of the class) using the main `train` function.

    # Arguments
        hdf5_file: The HDF5 file containing the preprocessed images and their respective annotations.
        body_backbone_CNN: Truncated version of a CNN which takes as input the region of the image comprising
            the person who’s feelings are to be estimated. One of `VGG16`, `VGG19`, `ResNet50` or `VGG16_Places365`.
        image_backbone_CNN: Truncated version of a CNN which takes as input the the entire image.
            One of `VGG16`, `VGG19`, `ResNet50` or `VGG16_Places365`.
        nb_of_epochs: Integer, total number of iterations on the data.
        weights_to_file: File name or full path for saving the weights of the current training process.
        modelCheckpoint_quantity: Quantity to monitor when saving the model after every epoch is enabled.
        earlyStopping_quantity: Quantity to monitor when stopping training when a monitored quantity has stopped improving is enabled.
        CSVLogger_filename: filename of the csv file, where the CSVLogger callback will stream epoch results to.

    # Raises
        ValueError: in case of invalid argument for `body_backbone_CNN`
            or invalid argument for `image_backbone_CNN`.
    """


    def __init__(self,
                 hdf5_file,
                 body_backbone_CNN,
                 image_backbone_CNN,
                 nb_of_epochs,
                 weights_to_file,
                 modelCheckpoint_quantity,
                 earlyStopping_quantity,
                 CSVLogger_filename):


        if not (body_backbone_CNN in {'VGG16', 'VGG19', 'ResNet50', 'VGG16_Places365'}):
            raise ValueError('The `body_backbone_CNN` argument should be either '
                             '`VGG16`, `VGG19`, `ResNet50` or `VGG16_Places365`. ')

        if not (image_backbone_CNN in {'VGG16', 'VGG19', 'ResNet50', 'VGG16_Places365'}):
            raise ValueError('The `image_backbone_CNN` argument should be either '
                             '`VGG16`, `VGG19`, `ResNet50` or `VGG16_Places365`. ')

        self.body_backbone_CNN = body_backbone_CNN
        self.image_backbone_CNN = image_backbone_CNN

        # -------------------------------------------------------------------------------- #
        #                             Construct EMOTIC model
        # -------------------------------------------------------------------------------- #

        body_inputs = Input(shape=(224, 224, 3), name='INPUT')
        image_inputs = Input(shape=(224, 224, 3), name='INPUT')

        # Body module
        if 'VGG16' == body_backbone_CNN:
            self.body_truncated_model = VGG16(include_top=False, weights='imagenet', input_tensor=body_inputs, pooling='avg')

        elif 'VGG19' == body_backbone_CNN:
            self.body_truncated_model = VGG19(include_top=False, weights='imagenet', input_tensor=body_inputs, pooling='avg')

        elif 'ResNet50' == body_backbone_CNN:
            tmp_model = ResNet50(include_top=False, weights='imagenet', input_tensor=body_inputs, pooling='avg')
            self.body_truncated_model = Model(inputs=tmp_model.input, outputs=tmp_model.get_layer('activation_48').output)

        elif 'VGG16_Places365' == body_backbone_CNN:
            self.body_truncated_model = VGG16_Places365(include_top=False, weights='places', input_tensor=body_inputs, pooling='avg')

        for layer in self.body_truncated_model.layers:
            layer.name = str("body-") + layer.name


        print('[INFO] The plain, body `' + body_backbone_CNN + '` pre-trained convnet was successfully initialised.')

        # Image module
        if 'VGG16' == image_backbone_CNN:
            self.image_truncated_model = VGG16(include_top=False, weights='imagenet', input_tensor=image_inputs, pooling='avg')

        elif 'VGG19' == image_backbone_CNN:
            self.image_truncated_model = VGG19(include_top=False, weights='imagenet', input_tensor=image_inputs, pooling='avg')

        elif 'ResNet50' == image_backbone_CNN:
            tmp_model = ResNet50(include_top=False, weights='imagenet',input_tensor=image_inputs, pooling='avg')
            self.image_truncated_model = Model(inputs=tmp_model.input, outputs=tmp_model.get_layer('activation_48').output)

        elif 'VGG16_Places365' == image_backbone_CNN:
            self.image_truncated_model = VGG16_Places365(include_top=False, weights='places', input_tensor=image_inputs, pooling='avg')

        for layer in self.image_truncated_model.layers:
            layer.name = str("image-") + layer.name

        print('[INFO] The plain, image `' + image_backbone_CNN + '` pre-trained convnet was successfully initialised.')

        # retrieve the ouputs
        body_plain_model_output = self.body_truncated_model.output
        image_plain_model_output = self.image_truncated_model.output


        # In case ResNet50 is selected we need to use a global average pooling layer to follow the process used for the othe CNNs.
        if 'ResNet50' == body_backbone_CNN:
            body_plain_model_output = GlobalAveragePooling2D(name='GAP')(body_plain_model_output)

        if 'ResNet50' == image_backbone_CNN:
            image_plain_model_output = GlobalAveragePooling2D(name='GAP')(image_plain_model_output)

        merged = concatenate([body_plain_model_output, image_plain_model_output])

        x = Dense(256, activation='relu', name='FC1', kernel_regularizer=regularizers.l2(0.01), kernel_initializer='random_normal')(merged)

        x = Dropout(0.5, name='DROPOUT')(x)

        vad_cont_prediction = Dense(units=3, kernel_initializer='random_normal', name='VAD')(x)

        # At model instantiation, you specify the two inputs and the output.
        self.model = Model(inputs=[body_inputs, image_inputs], outputs=vad_cont_prediction, name='EMOTIC-VAD-regression')

        print('[INFO] Randomly initialised classifier was successfully added on top of the merged modules.')

        print('[INFO] Number of trainable weights before freezing the conv. bases of the respective original models: '
              '' + str(len(self.model.trainable_weights)))

        # first: train only the top layers (which were randomly initialized)
        # i.e. freeze all convolutional layers of the preliminary base model
        for layer in self.body_truncated_model.layers:
            layer.trainable = False

        for layer in self.image_truncated_model.layers:
            layer.trainable = False

        print('[INFO] Number of trainable weights after freezing the conv. bases of the respective original models: '
              '' + str(len(self.model.trainable_weights)))

        # # reference https://github.com/keras-team/keras/issues/4735#issuecomment-267472549
        # self.class_weight = { 'VALENCE': {0: 36.00, 1: 36.00, 2: 12.00, 3: 5.14, 4: 2.25, 5: 1.00, 6: 1.89, 7: 2.57, 8: 12.00, 9: 36.00},
        #                       'AROUSAL': {0: 23.00, 1: 11.50, 2: 4.60, 3: 1.00, 4: 2.09, 5: 1.64, 6: 1.14, 7: 2.09, 8: 3.83, 9: 4.60},
        #                       'DOMINANCE': {0: 34.00, 1: 17.00, 2: 11.33, 3: 6.80, 4: 5.66, 5: 1.70, 6: 1.00, 7: 2.42, 8: 3.40, 9: 6.80}
        #                     }


        self.model.compile(optimizer=SGD(lr=1e-5, momentum=0.9),
                           # loss='mse',
                           loss = euclidean_distance_loss,
                           metrics=['mae','mse', rmse])

        # print ('[INFO] Metrics names: ',self.model.metrics_names )

        print('[INFO] End-to-end `EMOTIC-VAD-regression` model has been successfully compiled.')

        # -------------------------------------------------------------------------------- #
        #                                    Configurations
        # -------------------------------------------------------------------------------- #


        nb_train_samples = 23706
        nb_val_samples = 3332
        nb_test_samples = 7280

        train_generator_batch_size = 54
        val_generator_batch_size = 49
        test_generator_batch_size = 52

        self.steps_per_epoch = nb_train_samples // train_generator_batch_size
        self.validation_steps = nb_val_samples // val_generator_batch_size
        self.test_steps = nb_test_samples // test_generator_batch_size


        # -------------------------------------------------------------------------------- #
        #                                Read the HDF5 file
        # -------------------------------------------------------------------------------- #
        # open the hdf5 file
        hdf5_file = h5py.File(hdf5_file, "r")

        self.nb_train_data = hdf5_file["x_image_train"].shape[0]

        self.nb_val_data = hdf5_file["x_image_val"].shape[0]

        self.nb_test_data = hdf5_file["x_image_test"].shape[0]



        # -------------------------------------------------------------------------------- #
        #                         Instantiate the custom generators
        # -------------------------------------------------------------------------------- #

        print('[INFO] Setting up custom generators...')

        self.train_generator = custom_generator_single_output(hdf5_file=hdf5_file,
                                                              nb_data=self.nb_train_data,
                                                              batch_size=train_generator_batch_size,
                                                              mode='train')

        self.val_generator = custom_generator_single_output(hdf5_file=hdf5_file,
                                                            nb_data=self.nb_val_data,
                                                            batch_size=val_generator_batch_size,
                                                            mode='val')

        self.test_generator = custom_generator_single_output(hdf5_file=hdf5_file,
                                                             nb_data=self.nb_test_data,
                                                             batch_size=test_generator_batch_size,
                                                             mode='test')




        # -------------------------------------------------------------------------------- #
        #                                Usage of callbacks
        # -------------------------------------------------------------------------------- #

        self.weights_to_file = weights_to_file
        self.nb_of_epochs = nb_of_epochs

        # CSVLogger
        model_log = 'trained_models/logs/' + CSVLogger_filename
        csv_logger = CSVLogger(model_log, append=True, separator=',')


        # ModelCheckpoint
        checkpointer = ModelCheckpoint(filepath=weights_to_file,
                                       monitor=modelCheckpoint_quantity,
                                       verbose=1,
                                       save_best_only=True,
                                       mode='auto',
                                       period=1,
                                       save_weights_only=True)

        early_stop = EarlyStopping(monitor=earlyStopping_quantity, patience=5, mode='auto')

        self.callbacks_list = [checkpointer, early_stop, csv_logger]




    def train(self):
        """Trains the EMOTIC model for a given number of epochs (iterations on a dataset).

        """

        self.model.summary()

        print('[INFO] Start training the end-to-end EMOTIC model...')

        start_time = time.time()

        history = self.model.fit_generator(self.train_generator,
                                           epochs=self.nb_of_epochs,
                                           steps_per_epoch=self.steps_per_epoch,
                                           validation_data=self.val_generator,
                                           validation_steps=self.validation_steps,
                                           callbacks=self.callbacks_list,
                                           # class_weight= self.class_weight
                                           )

        end_time = time.time()
        print("[INFO] It took {} to train the end-to-end EMOTIC model".format(hms_string(end_time - start_time)))

        print('[INFO] Saved trained model as: %s ' % self.weights_to_file)





"""Python utility which reads image names (locations on disk) from a csv file,
    loads them using the basic set of tools for image data provided by Keras (keras/preprocessing/image.py)
    and saves them in a numpy array.
"""

import numpy as np
import pandas
from utils.generic_utils import progress, imagenet_preprocess_input, places_preprocess_input
from keras.preprocessing import image
from keras.applications.vgg16 import preprocess_input
from keras.applications.resnet50 import preprocess_input


# reference https://stackoverflow.com/questions/12984426/python-pil-ioerror-image-file-truncated-with-big-images
from PIL import ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True


def save_x_train_to_numpy(base_img_dir,
                          base_csv_dir,
                          input_size,
                          input_img_mode,
                          mode='train',
                          to_file = 'numpy_annotations/x_train'
                          ):
    """ Saves images loaded from a CSV to numpy array.

        # Arguments
            base_img_dir: the directory where the raw images are stored.
            In our setup, we:
            - created train/ val/ and test/ subfolders inside EMOTIC_database/

            base_csv_dir: the directory where the CSV files are stored.
            input_size: the default input size for the model (ref https://keras.io/applications/).
                All models have input size of 224x224 except Xception,InceptionV3 and InceptionResNetV2 which have input size of 299x299.
            input_img_mode: one of `body` (cropped images) or `body` (entire images)
            mode: one of `train` (train set), `val` (validation set)
                or `test` (test set).
            to_file: the name or path of the numpy array where the images will be saved.
        """


    # Load CSV File With Pandas
    csv_name = base_csv_dir + mode + '.csv'
    csv_file = pandas.read_csv(csv_name)

    if mode == 'train':
        nb_samples = 23706
    elif mode == 'val':
        nb_samples = 3332
    elif mode == 'test':
        nb_samples = 7280

    field_number = 0

    # pre-allocating the data array, and then loading the data directly into it
    # ref: https://hjweide.github.io/efficient-image-loading
    data = np.empty((nb_samples, input_size, input_size, 3), dtype=np.uint8)

    for img_name in csv_file.filename:
        progress(field_number, nb_samples)

        img_name = base_img_dir + img_name
        img = image.load_img(img_name, target_size=(input_size, input_size)) # load an image from file
        x = image.img_to_array(img) # this is a Numpy array with shape (input_size, input_size, 3)
        x = np.expand_dims(x, axis=0) # this is a Numpy array with shape (1, input_size, input_size, 3)

        # prepare the image for either ImageNet-based or Places-based models
        if input_img_mode == 'body':
            x = imagenet_preprocess_input(x)
        elif input_img_mode == 'image':
            x = places_preprocess_input(x)


        # x = (x / 255.).astype(np.float32)

        # scaling the RGB values to a 0-1.0 range.
        # reference https://www.linkedin.com/pulse/keras-image-preprocessing-scaling-pixels-training-adwin-jahn/
        # x /= 255
        data[field_number, ...] = x

        field_number += 1
        if field_number > nb_samples - 1:
            break

    np.save(to_file, data)

    return data


if __name__ == '__main__':

    # one of `train` (train set), `val` (validation set) or `test` (test set).
    mode = 'test'

    # one of `body` or `image`
    input_img_mode = 'image'





    if input_img_mode == 'body':
        base_img_dir = '/home/sandbox/Desktop/EMOTIC_resources/raw_refined_images/cropped_imgs/'
        to_file = 'x_body_'+mode
    elif input_img_mode == 'image':
        base_img_dir = '/home/sandbox/Desktop/EMOTIC_resources/raw_refined_images/entire_multiple_imgs/'
        to_file = 'x_image_' + mode



    x_train = save_x_train_to_numpy(base_img_dir=base_img_dir,
                                    base_csv_dir='/home/sandbox/Desktop/EMOTIC_resources/VAD-classification/CSV/',
                                    input_size=224,
                                    input_img_mode = input_img_mode,
                                    mode=mode,
                                    to_file=to_file)



from hdf5_controller import Controller



hdf5_file_name = 'EMOTIC-VAD-Classification-rescale.hdf5'

train_csv_file_path ='/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/CSV/train.csv'
val_csv_file_path ='/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/CSV/val.csv'
test_csv_file_path ='/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/CSV/test.csv'

cropped_imgs_dir ='/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/cropped_imgs/'
entire_imgs_dir = '/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/entire_multiple_imgs/'
main_numpy_dir ='/home/gkallia/git/emotic-VAD-classification/dataset/EMOTIC_resources/numpy_matrices/'

controller = Controller(hdf5_file=hdf5_file_name,
                        train_csv_file_path=train_csv_file_path,
                        val_csv_file_path=val_csv_file_path,
                        test_csv_file_path=test_csv_file_path,
                        cropped_imgs_dir=cropped_imgs_dir,
                        entire_imgs_dir=entire_imgs_dir,
                        main_numpy_dir=main_numpy_dir)


create_hdf5 = controller.create_hdf5_VAD_regression(dataset='EMOTIC', input_size=224)



# -*- coding: utf-8 -*-
""" HRA_Transfer_CNN_Manager is a wrapper class which 'encapsulates' the functionalities needed for preparing (class instantiation)
    and training different CNNs (`train_model`) on the HRA dataset with 2 classes.
"""

from __future__ import print_function
import os
import sys
import math
import numpy as np
import os.path
import time
import datetime
import h5py
from keras.callbacks import ModelCheckpoint, EarlyStopping, CSVLogger
from keras.preprocessing.image import ImageDataGenerator

from applications.hra_vgg16 import HRA_VGG16
from applications.hra_vgg19 import HRA_VGG19
from applications.hra_resnet50 import HRA_ResNet50
from applications.hra_vgg16_places365 import HRA_VGG16_Places365

from utils.generic_utils import hms_string


class HRA_Transfer_CNN_Manager():
    """Loads the parameters needed for the training process on class instantiation
        & starts the training process.

        # Arguments
            violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
                or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
            train_mode: String to declare the train mode of the model (how many layers will be frozen during training).
                - `feature_extraction` taking the convolutional base of a previously-trained network,
                    running the new data through it, and training a new classifier on top of the output.
                - `fine_tuning` unfreezing a few of the top layers of a frozen conv. base used for feature extraction,
                    and jointly training both the newly added part of the model and these top layers.
            pre_trained_model: One of `vgg16`, `vgg19`, `resnet50` or `vgg16_places365`.
            nb_of_conv_layers_to_fine_tune: integer to indicate the number of convolutional layers to fine-tune.
                One of `None` (indicates feature extraction mode), `1`, `2` or `3`.
            weights_to_file: File name or full path for saving the weights of the current training process.
            first_phase_trained_weights: Weights of an already trained feature extraction model.
                Only relevant when using `fine_tuning` as train_mode after `feature_extraction` weights have been saved.
            nb_of_epochs: Integer, total number of iterations on the data.
            modelCheckpoint_quantity: Quantity to monitor when saving the model after every epoch is enabled.
            earlyStopping_quantity: Quantity to monitor when stopping training when a monitored quantity has stopped improving is enabled.
            CSVLogger_filename: filename of the csv file, where the CSVLogger callback will stream epoch results to.

        # Raises
            ValueError: in case of invalid argument for `nb_of_conv_layers_to_fine_tune`
                or invalid argument for `first_phase_trained_weights`.
    """

    def __init__(self,
                 violation_class,
                 train_mode,
                 pre_trained_model,
                 nb_of_conv_layers_to_fine_tune,
                 weights_to_file,
                 first_phase_trained_weights,
                 nb_of_epochs,
                 modelCheckpoint_quantity,
                 earlyStopping_quantity,
                 CSVLogger_filename,
                 ):


        # extra check for the case when fine-tuning is selected without providing the correct first_phase_trained_weights.
        if  nb_of_conv_layers_to_fine_tune in {1, 2, 3} and first_phase_trained_weights is None:
            raise ValueError('The `first_phase_trained_weights` argument can be set to None only when '
                             '`nb_of_conv_layers_to_fine_tune` is None (feature extraction).'
                             'When `nb_of_conv_layers_to_fine_tune` is either 1 or 2, '
                             'the weights of an already trained feature extraction model must be saved prior to fine-tuning the model.')

        # Base directory for saving the trained models
        self.trained_models_dir = '/home/gkallia/git/Human-Rights-Violations-Conceptron/trained_models'
        self.feature_extraction_dir = os.path.join(self.trained_models_dir, 'feature_extraction/')
        self.fine_tuning_dir = os.path.join(self.trained_models_dir, 'fine_tuning/')
        self.logs_dir = os.path.join(self.trained_models_dir, 'logs/')

        if violation_class == 'cl':
            self.train_dir = os.path.join('/home/gkallia/git/Human-Rights-Violations-Conceptron/datasets/Two-class-HRV/ChildLabour', 'train')
            self.val_dir = os.path.join('/home/gkallia/git/Human-Rights-Violations-Conceptron/datasets/Two-class-HRV/ChildLabour', 'val')

        elif violation_class == 'dp':
            self.train_dir = os.path.join('/home/gkallia/git/Human-Rights-Violations-Conceptron/datasets/Two-class-HRV/DisplacedPopulations', 'train')
            self.val_dir = os.path.join('/home/gkallia/git/Human-Rights-Violations-Conceptron/datasets/Two-class-HRV/DisplacedPopulations', 'val')



        # Augmentation configuration with only rescaling.
        # Rescale is a value by which we will multiply the data before any other processing.
        # Our original images consist in RGB coefficients in the 0-255, but such values would
        # be too high for our models to process (given a typical learning rate),
        # so we target values between 0 and 1 instead by scaling with a 1/255. factor.
        datagen = ImageDataGenerator(rescale=1. / 255)

        img_width, img_height = 224, 224

        self.train_batch_size = 21
        self.val_batch_size = 10


        print('[INFO] Setting up image data generators...')

        self.train_generator = datagen.flow_from_directory(self.train_dir, target_size=(img_width, img_height),
                                                           class_mode='categorical',
                                                           shuffle=False,
                                                           batch_size=self.train_batch_size)

        self.val_generator = datagen.flow_from_directory(self.val_dir, target_size=(img_width, img_height),
                                                         class_mode='categorical',
                                                         shuffle=False,
                                                         batch_size=self.val_batch_size)


        num_classes = len(self.train_generator.class_indices)

        print('[INFO] Number of classes: ', num_classes)

        self.nb_train_samples = len(self.train_generator.filenames)
        # train_labels = self.train_generator.classes
        # self.train_labels = to_categorical(train_labels, num_classes=num_classes)
        # self.predict_size_train = int(math.ceil(self.nb_train_samples / self.train_batch_size))

        print ('[INFO] Number of train samples: ', self.nb_train_samples)

        # print('[INFO] Predict size train: ', self.predict_size_train)

        # save the class indices to use use later in predictions
        # np.save('class_indices.npy', self.train_generator.class_indices)


        self.nb_val_samples = len(self.val_generator.filenames)
        # val_labels = self.val_generator.classes
        # self.val_labels = to_categorical(val_labels, num_classes=num_classes)
        # self.predict_size_test = int(math.ceil(self.nb_val_samples / self.val_batch_size))

        print ('[INFO] Number of test samples: ', self.nb_val_samples)
        # print('[INFO] Predict size test: ', self.predict_size_test)

        self.steps_per_epoch = self.nb_train_samples // self.train_batch_size
        self.val_steps = self.nb_val_samples // self.val_batch_size



        # -------------------------------------------------------------------------------- #
        #                                Usage of callbacks
        # -------------------------------------------------------------------------------- #

        self.train_mode = train_mode
        self.pre_trained_model = pre_trained_model
        self.nb_of_conv_layers_to_fine_tune = nb_of_conv_layers_to_fine_tune
        self.weights_to_file = weights_to_file
        self.first_phase_trained_weights = first_phase_trained_weights
        self.nb_of_epochs = nb_of_epochs
        # self.modelCheckpoint_quantity = modelCheckpoint_quantity
        # self.earlyStopping_quantity = earlyStopping_quantity
        # self.CSVLogger_filename = CSVLogger_filename

        # self.steps_per_epoch = self.nb_train_samples // self.train_batch_size
        #
        #
        # self.val_steps = self.nb_val_samples // self.val_batch_size



        # CSVLogger
        model_log = 'trained_models/logs/' + CSVLogger_filename
        csv_logger = CSVLogger(model_log, append=True, separator=',')


        # ModelCheckpoint
        checkpointer = ModelCheckpoint(filepath=weights_to_file,
                                       monitor=modelCheckpoint_quantity,
                                       verbose=1,
                                       save_best_only=True,
                                       mode='auto',
                                       period=1,
                                       save_weights_only=True)

        early_stop = EarlyStopping(monitor=earlyStopping_quantity, patience=5, mode='auto')

        self.callbacks_list = [checkpointer, early_stop, csv_logger]


    def train(self):
        """Loads the selected model & starts the training process.
        """

        if self.pre_trained_model == 'vgg16':

            print('[INFO] Instantiating HRA-2CLASS-VGG16...')

            if self.train_mode == 'feature_extraction':

                model = HRA_VGG16(include_top=True, weights=None,
                                  input_tensor=None, input_shape=None,
                                  nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                  first_phase_trained_weights=None,
                                  verbose=1)
            else:

                if os.path.isfile(self.first_phase_trained_weights) is False:
                    raise IOError("No such weights file: `" + self.first_phase_trained_weights + "`. ")

                model = HRA_VGG16(include_top=True, weights=None,
                                  input_tensor=None, input_shape=None,
                                  nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                  first_phase_trained_weights=self.first_phase_trained_weights,
                                  verbose=1)

            print('[INFO] HRA-2CLASS-VGG16 model loaded')



        elif self.pre_trained_model == 'vgg19':

            print('[INFO] Instantiating HRA-2CLASS-VGG19...')

            if self.train_mode == 'feature_extraction':

                model = HRA_VGG19(include_top=True, weights=None,
                                  input_tensor=None, input_shape=None,
                                  nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                  first_phase_trained_weights=None,
                                  verbose=1)
            else:

                if os.path.isfile(self.first_phase_trained_weights) is False:
                    raise IOError("No such weights file: `" + self.first_phase_trained_weights + "`. ")

                model = HRA_VGG19(include_top=True, weights=None,
                                  input_tensor=None, input_shape=None,
                                  nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                  first_phase_trained_weights=self.first_phase_trained_weights,
                                  verbose=1)

            print('[INFO] HRA-2CLASS-VGG19 model loaded')




        elif self.pre_trained_model == 'resnet50':

            print('[INFO] Instantiating HRA-2CLASS-ResNet50...')

            if self.train_mode == 'feature_extraction':

                model = HRA_ResNet50(include_top=True, weights=None,
                                     input_tensor=None, input_shape=None,
                                     nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                     first_phase_trained_weights=None,
                                     verbose=1)
            else:

                if os.path.isfile(self.first_phase_trained_weights) is False:
                    raise IOError("No such weights file: `" + self.first_phase_trained_weights + "`. ")

                model = HRA_ResNet50(include_top=True, weights=None,
                                     input_tensor=None, input_shape=None,
                                     nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                     first_phase_trained_weights=self.first_phase_trained_weights,
                                     verbose=1)

            print('[INFO] HRA-2CLASS-ResNet50 model loaded')




        elif self.pre_trained_model == 'vgg16_places365':

            print('[INFO] Instantiating HRA-2CLASS-VGG16_Places365...')

            if self.train_mode == 'feature_extraction':

                model = HRA_VGG16_Places365(include_top=True, weights=None,
                                            input_tensor=None, input_shape=None,
                                            nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                            first_phase_trained_weights=None,
                                            verbose=1)
            else:

                if os.path.isfile(self.first_phase_trained_weights) is False:
                    raise IOError("No such weights file: `" + self.first_phase_trained_weights + "`. ")

                model = HRA_VGG16_Places365(include_top=True, weights=None,
                                            input_tensor=None, input_shape=None,
                                            nb_of_conv_layers_to_fine_tune=self.nb_of_conv_layers_to_fine_tune,
                                            first_phase_trained_weights=self.first_phase_trained_weights,
                                            verbose=1)

            print('[INFO] HRA-2CLASS-VGG16-Places365 model loaded')



        # Finally start fitting the dataset

        if self.train_mode == 'feature_extraction':
            print('[INFO] Start training the randomly initialised classifier on top of the pre-trained conv. base...')

            start_time = time.time()

            history = model.fit_generator(self.train_generator,
                                          epochs=self.nb_of_epochs,
                                          steps_per_epoch=self.steps_per_epoch,
                                          validation_data=self.val_generator,
                                          validation_steps=self.val_steps,
                                          callbacks=self.callbacks_list)

            end_time = time.time()
            print("[INFO] It took {} to train the randomly initialised classifier on top of the pre-trained conv. base".format(
                hms_string(end_time - start_time)))

            print('[INFO] Saved trained model as: %s ' % self.weights_to_file)

        else:
            print('[INFO] Start fine-tuning the model...')

            start_time = time.time()

            history = model.fit_generator(self.train_generator,
                                          epochs=self.nb_of_epochs,
                                          steps_per_epoch=self.steps_per_epoch,
                                          validation_data=self.val_generator,
                                          validation_steps=self.val_steps,
                                          callbacks=self.callbacks_list)

            end_time = time.time()
            print("[INFO] It took {} to fine-tune the top layers of the frozen conv. base".format(
                hms_string(end_time - start_time)))

            print('[INFO] Saved trained model as: %s ' % self.weights_to_file)




'''
Utilities for 2D object detection related to answering the following questions:
1. Given an image size and bounding boxes, which bounding boxes meet certain
   requirements with respect to the image size?
2. Given an image size and bounding boxes, is an image of that size valid with
   respect to the bounding boxes according to certain requirements?

Copyright (C) 2018 Pierluigi Ferrari

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from __future__ import division

import numpy as np

from engine.object_detection_branch.single_shot_detector.bounding_box_utils.bounding_box_utils import iou


class BoundGenerator:
    '''
    Generates pairs of floating point values that represent lower and upper bounds
    from a given sample space.
    '''
    def __init__(self,
                 sample_space=((0.1, None),
                               (0.3, None),
                               (0.5, None),
                               (0.7, None),
                               (0.9, None),
                               (None, None)),
                 weights=None):
        '''
        Arguments:
            sample_space (list or tuple): A list, tuple, or array-like object of shape
                `(n, 2)` that contains `n` samples to choose from, where each sample
                is a 2-tuple of scalars and/or `None` values.
            weights (list or tuple, optional): A list or tuple representing the distribution
                over the sample space. If `None`, a uniform distribution will be assumed.
        '''

        if (not (weights is None)) and len(weights) != len(sample_space):
            raise ValueError("`weights` must either be `None` for uniform distribution or have the same length as `sample_space`.")

        self.sample_space = []
        for bound_pair in sample_space:
            if len(bound_pair) != 2:
                raise ValueError("All elements of the sample space must be 2-tuples.")
            bound_pair = list(bound_pair)
            if bound_pair[0] is None: bound_pair[0] = 0.0
            if bound_pair[1] is None: bound_pair[1] = 1.0
            if bound_pair[0] > bound_pair[1]:
                raise ValueError("For all sample space elements, the lower bound cannot be greater than the upper bound.")
            self.sample_space.append(bound_pair)

        self.sample_space_size = len(self.sample_space)

        if weights is None:
            self.weights = [1.0/self.sample_space_size] * self.sample_space_size
        else:
            self.weights = weights

    def __call__(self):
        '''
        Returns:
            An item of the sample space, i.e. a 2-tuple of scalars.
        '''
        i = np.random.choice(self.sample_space_size, p=self.weights)
        return self.sample_space[i]

class BoxFilter:
    '''
    Returns all bounding boxes that are valid with respect to a the defined criteria.
    '''

    def __init__(self,
                 check_overlap=True,
                 check_min_area=True,
                 check_degenerate=True,
                 overlap_criterion='center_point',
                 overlap_bounds=(0.3, 1.0),
                 min_area=16,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4},
                 border_pixels='half'):
        '''
        Arguments:
            check_overlap (bool, optional): Whether or not to enforce the overlap requirements defined by
                `overlap_criterion` and `overlap_bounds`. Sometimes you might want to use the box filter only
                to enforce a certain minimum area for all boxes (see next argument), in such cases you can
                turn the overlap requirements off.
            check_min_area (bool, optional): Whether or not to enforce the minimum area requirement defined
                by `min_area`. If `True`, any boxes that have an area (in pixels) that is smaller than `min_area`
                will be removed from the labels of an image. Bounding boxes below a certain area aren't useful
                training examples. An object that takes up only, say, 5 pixels in an image is probably not
                recognizable anymore, neither for a human, nor for an object detection model. It makes sense
                to remove such boxes.
            check_degenerate (bool, optional): Whether or not to check for and remove degenerate bounding boxes.
                Degenerate bounding boxes are boxes that have `xmax <= xmin` and/or `ymax <= ymin`. In particular,
                boxes with a width and/or height of zero are degenerate. It is obviously important to filter out
                such boxes, so you should only set this option to `False` if you are certain that degenerate
                boxes are not possible in your data and processing chain.
            overlap_criterion (str, optional): Can be either of 'center_point', 'iou', or 'area'. Determines
                which boxes are considered valid with respect to a given image. If set to 'center_point',
                a given bounding box is considered valid if its center point lies within the image.
                If set to 'area', a given bounding box is considered valid if the quotient of its intersection
                area with the image and its own area is within the given `overlap_bounds`. If set to 'iou', a given
                bounding box is considered valid if its IoU with the image is within the given `overlap_bounds`.
            overlap_bounds (list or BoundGenerator, optional): Only relevant if `overlap_criterion` is 'area' or 'iou'.
                Determines the lower and upper bounds for `overlap_criterion`. Can be either a 2-tuple of scalars
                representing a lower bound and an upper bound, or a `BoundGenerator` object, which provides
                the possibility to generate bounds randomly.
            min_area (int, optional): Only relevant if `check_min_area` is `True`. Defines the minimum area in
                pixels that a bounding box must have in order to be valid. Boxes with an area smaller than this
                will be removed.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
            border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
                Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
                to the boxes. If 'exclude', the border pixels do not belong to the boxes.
                If 'half', then one of each of the two horizontal and vertical borders belong
                to the boxex, but not the other.
        '''
        if not isinstance(overlap_bounds, (list, tuple, BoundGenerator)):
            raise ValueError("`overlap_bounds` must be either a 2-tuple of scalars or a `BoundGenerator` object.")
        if isinstance(overlap_bounds, (list, tuple)) and (overlap_bounds[0] > overlap_bounds[1]):
            raise ValueError("The lower bound must not be greater than the upper bound.")
        if not (overlap_criterion in {'iou', 'area', 'center_point'}):
            raise ValueError("`overlap_criterion` must be one of 'iou', 'area', or 'center_point'.")
        self.overlap_criterion = overlap_criterion
        self.overlap_bounds = overlap_bounds
        self.min_area = min_area
        self.check_overlap = check_overlap
        self.check_min_area = check_min_area
        self.check_degenerate = check_degenerate
        self.labels_format = labels_format
        self.border_pixels = border_pixels

    def __call__(self,
                 labels,
                 image_height=None,
                 image_width=None):
        '''
        Arguments:
            labels (array): The labels to be filtered. This is an array with shape `(m,n)`, where
                `m` is the number of bounding boxes and `n` is the number of elements that defines
                each bounding box (box coordinates, class ID, etc.). The box coordinates are expected
                to be in the image's coordinate system.
            image_height (int): Only relevant if `check_overlap == True`. The height of the image
                (in pixels) to compare the box coordinates to.
            image_width (int): `check_overlap == True`. The width of the image (in pixels) to compare
                the box coordinates to.

        Returns:
            An array containing the labels of all boxes that are valid.
        '''

        labels = np.copy(labels)

        xmin = self.labels_format['xmin']
        ymin = self.labels_format['ymin']
        xmax = self.labels_format['xmax']
        ymax = self.labels_format['ymax']

        # Record the boxes that pass all checks here.
        requirements_met = np.ones(shape=labels.shape[0], dtype=np.bool)

        if self.check_degenerate:

            non_degenerate = (labels[:,xmax] > labels[:,xmin]) * (labels[:,ymax] > labels[:,ymin])
            requirements_met *= non_degenerate

        if self.check_min_area:

            min_area_met = (labels[:,xmax] - labels[:,xmin]) * (labels[:,ymax] - labels[:,ymin]) >= self.min_area
            requirements_met *= min_area_met

        if self.check_overlap:

            # Get the lower and upper bounds.
            if isinstance(self.overlap_bounds, BoundGenerator):
                lower, upper = self.overlap_bounds()
            else:
                lower, upper = self.overlap_bounds

            # Compute which boxes are valid.

            if self.overlap_criterion == 'iou':
                # Compute the patch coordinates.
                image_coords = np.array([0, 0, image_width, image_height])
                # Compute the IoU between the patch and all of the ground truth boxes.
                image_boxes_iou = iou(image_coords, labels[:, [xmin, ymin, xmax, ymax]], coords='corners', mode='element-wise', border_pixels=self.border_pixels)
                requirements_met *= (image_boxes_iou > lower) * (image_boxes_iou <= upper)

            elif self.overlap_criterion == 'area':
                if self.border_pixels == 'half':
                    d = 0
                elif self.border_pixels == 'include':
                    d = 1 # If border pixels are supposed to belong to the bounding boxes, we have to add one pixel to any difference `xmax - xmin` or `ymax - ymin`.
                elif self.border_pixels == 'exclude':
                    d = -1 # If border pixels are not supposed to belong to the bounding boxes, we have to subtract one pixel from any difference `xmax - xmin` or `ymax - ymin`.
                # Compute the areas of the boxes.
                box_areas = (labels[:,xmax] - labels[:,xmin] + d) * (labels[:,ymax] - labels[:,ymin] + d)
                # Compute the intersection area between the patch and all of the ground truth boxes.
                clipped_boxes = np.copy(labels)
                clipped_boxes[:,[ymin,ymax]] = np.clip(labels[:,[ymin,ymax]], a_min=0, a_max=image_height-1)
                clipped_boxes[:,[xmin,xmax]] = np.clip(labels[:,[xmin,xmax]], a_min=0, a_max=image_width-1)
                intersection_areas = (clipped_boxes[:,xmax] - clipped_boxes[:,xmin] + d) * (clipped_boxes[:,ymax] - clipped_boxes[:,ymin] + d) # +1 because the border pixels belong to the box areas.
                # Check which boxes meet the overlap requirements.
                if lower == 0.0:
                    mask_lower = intersection_areas > lower * box_areas # If `self.lower == 0`, we want to make sure that boxes with area 0 don't count, hence the ">" sign instead of the ">=" sign.
                else:
                    mask_lower = intersection_areas >= lower * box_areas # Especially for the case `self.lower == 1` we want the ">=" sign, otherwise no boxes would count at all.
                mask_upper = intersection_areas <= upper * box_areas
                requirements_met *= mask_lower * mask_upper

            elif self.overlap_criterion == 'center_point':
                # Compute the center points of the boxes.
                cy = (labels[:,ymin] + labels[:,ymax]) / 2
                cx = (labels[:,xmin] + labels[:,xmax]) / 2
                # Check which of the boxes have center points within the cropped patch remove those that don't.
                requirements_met *= (cy >= 0.0) * (cy <= image_height-1) * (cx >= 0.0) * (cx <= image_width-1)

        return labels[requirements_met]

class ImageValidator:
    '''
    Returns `True` if a given minimum number of bounding boxes meets given overlap
    requirements with an image of a given height and width.
    '''

    def __init__(self,
                 overlap_criterion='center_point',
                 bounds=(0.3, 1.0),
                 n_boxes_min=1,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4},
                 border_pixels='half'):
        '''
        Arguments:
            overlap_criterion (str, optional): Can be either of 'center_point', 'iou', or 'area'. Determines
                which boxes are considered valid with respect to a given image. If set to 'center_point',
                a given bounding box is considered valid if its center point lies within the image.
                If set to 'area', a given bounding box is considered valid if the quotient of its intersection
                area with the image and its own area is within `lower` and `upper`. If set to 'iou', a given
                bounding box is considered valid if its IoU with the image is within `lower` and `upper`.
            bounds (list or BoundGenerator, optional): Only relevant if `overlap_criterion` is 'area' or 'iou'.
                Determines the lower and upper bounds for `overlap_criterion`. Can be either a 2-tuple of scalars
                representing a lower bound and an upper bound, or a `BoundGenerator` object, which provides
                the possibility to generate bounds randomly.
            n_boxes_min (int or str, optional): Either a non-negative integer or the string 'all'.
                Determines the minimum number of boxes that must meet the `overlap_criterion` with respect to
                an image of the given height and width in order for the image to be a valid image.
                If set to 'all', an image is considered valid if all given boxes meet the `overlap_criterion`.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
            border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
                Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
                to the boxes. If 'exclude', the border pixels do not belong to the boxes.
                If 'half', then one of each of the two horizontal and vertical borders belong
                to the boxex, but not the other.
        '''
        if not ((isinstance(n_boxes_min, int) and n_boxes_min > 0) or n_boxes_min == 'all'):
            raise ValueError("`n_boxes_min` must be a positive integer or 'all'.")
        self.overlap_criterion = overlap_criterion
        self.bounds = bounds
        self.n_boxes_min = n_boxes_min
        self.labels_format = labels_format
        self.border_pixels = border_pixels
        self.box_filter = BoxFilter(check_overlap=True,
                                    check_min_area=False,
                                    check_degenerate=False,
                                    overlap_criterion=self.overlap_criterion,
                                    overlap_bounds=self.bounds,
                                    labels_format=self.labels_format,
                                    border_pixels=self.border_pixels)

    def __call__(self,
                 labels,
                 image_height,
                 image_width):
        '''
        Arguments:
            labels (array): The labels to be tested. The box coordinates are expected
                to be in the image's coordinate system.
            image_height (int): The height of the image to compare the box coordinates to.
            image_width (int): The width of the image to compare the box coordinates to.

        Returns:
            A boolean indicating whether an imgae of the given height and width is
            valid with respect to the given bounding boxes.
        '''

        self.box_filter.overlap_bounds = self.bounds
        self.box_filter.labels_format = self.labels_format

        # Get all boxes that meet the overlap requirements.
        valid_labels = self.box_filter(labels=labels,
                                       image_height=image_height,
                                       image_width=image_width)

        # Check whether enough boxes meet the requirements.
        if isinstance(self.n_boxes_min, int):
            # The image is valid if at least `self.n_boxes_min` ground truth boxes meet the requirements.
            if len(valid_labels) >= self.n_boxes_min:
                return True
            else:
                return False
        elif self.n_boxes_min == 'all':
            # The image is valid if all ground truth boxes meet the requirements.
            if len(valid_labels) == len(labels):
                return True
            else:
                return False



#!/usr/bin/env python

"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import numpy as np
import os
import sys

import cv2

# Allow relative imports when being executed as script.
if __name__ == "__main__" and __package__ is None:
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..'))
    import keras_retinanet.bin  # noqa: F401
    __package__ = "keras_retinanet.bin"

# Change these to absolute imports if you copy this script outside the keras_retinanet package.
from ..preprocessing.pascal_voc import PascalVocGenerator
from ..preprocessing.csv_generator import CSVGenerator
from ..preprocessing.kitti import KittiGenerator
from ..preprocessing.open_images import OpenImagesGenerator
from ..utils.transform import random_transform_generator
from ..utils.visualization import draw_annotations, draw_boxes


def create_generator(args):
    """ Create the data generators.

    Args:
        args: parseargs arguments object.
    """
    # create random transform generator for augmenting training data
    transform_generator = random_transform_generator(
        min_rotation=-0.1,
        max_rotation=0.1,
        min_translation=(-0.1, -0.1),
        max_translation=(0.1, 0.1),
        min_shear=-0.1,
        max_shear=0.1,
        min_scaling=(0.9, 0.9),
        max_scaling=(1.1, 1.1),
        flip_x_chance=0.5,
        flip_y_chance=0.5,
    )

    if args.dataset_type == 'coco':
        # import here to prevent unnecessary dependency on cocoapi
        from ..preprocessing.coco import CocoGenerator

        generator = CocoGenerator(
            args.coco_path,
            args.coco_set,
            transform_generator=transform_generator
        )
    elif args.dataset_type == 'pascal':
        generator = PascalVocGenerator(
            args.pascal_path,
            args.pascal_set,
            transform_generator=transform_generator
        )
    elif args.dataset_type == 'csv':
        generator = CSVGenerator(
            args.annotations,
            args.classes,
            transform_generator=transform_generator
        )
    elif args.dataset_type == 'oid':
        generator = OpenImagesGenerator(
            args.main_dir,
            subset=args.subset,
            version=args.version,
            labels_filter=args.labels_filter,
            fixed_labels=args.fixed_labels,
            annotation_cache_dir=args.annotation_cache_dir,
            transform_generator=transform_generator
        )
    elif args.dataset_type == 'kitti':
        generator = KittiGenerator(
            args.kitti_path,
            subset=args.subset,
            transform_generator=transform_generator
        )
    else:
        raise ValueError('Invalid data type received: {}'.format(args.dataset_type))

    return generator


def parse_args(args):
    """ Parse the arguments.
    """
    parser     = argparse.ArgumentParser(description='Debug script for a RetinaNet network.')
    subparsers = parser.add_subparsers(help='Arguments for specific dataset types.', dest='dataset_type')
    subparsers.required = True

    coco_parser = subparsers.add_parser('coco')
    coco_parser.add_argument('coco_path',  help='Path to dataset directory (ie. /tmp/COCO).')
    coco_parser.add_argument('--coco-set', help='Name of the set to show (defaults to val2017).', default='val2017')

    pascal_parser = subparsers.add_parser('pascal')
    pascal_parser.add_argument('pascal_path', help='Path to dataset directory (ie. /tmp/VOCdevkit).')
    pascal_parser.add_argument('--pascal-set',  help='Name of the set to show (defaults to test).', default='test')

    kitti_parser = subparsers.add_parser('kitti')
    kitti_parser.add_argument('kitti_path', help='Path to dataset directory (ie. /tmp/kitti).')
    kitti_parser.add_argument('subset', help='Argument for loading a subset from train/val.')

    def csv_list(string):
        return string.split(',')

    oid_parser = subparsers.add_parser('oid')
    oid_parser.add_argument('main_dir', help='Path to dataset directory.')
    oid_parser.add_argument('subset', help='Argument for loading a subset from train/validation/test.')
    oid_parser.add_argument('--version',  help='The current dataset version is v4.', default='v4')
    oid_parser.add_argument('--labels-filter',  help='A list of labels to filter.', type=csv_list, default=None)
    oid_parser.add_argument('--annotation-cache-dir', help='Path to store annotation cache.', default='.')
    oid_parser.add_argument('--fixed-labels', help='Use the exact specified labels.', default=False)

    csv_parser = subparsers.add_parser('csv')
    csv_parser.add_argument('annotations', help='Path to CSV file containing annotations for evaluation.')
    csv_parser.add_argument('classes',     help='Path to a CSV file containing class label mapping.')

    parser.add_argument('-l', '--loop', help='Loop forever, even if the dataset is exhausted.', action='store_true')
    parser.add_argument('--no-resize', help='Disable image resizing.', dest='resize', action='store_false')
    parser.add_argument('--anchors', help='Show positive anchors on the image.', action='store_true')
    parser.add_argument('--annotations', help='Show annotations on the image. Green annotations have anchors, red annotations don\'t and therefore don\'t contribute to training.', action='store_true')
    parser.add_argument('--random-transform', help='Randomly transform image and annotations.', action='store_true')

    return parser.parse_args(args)


def run(generator, args):
    """ Main loop.

    Args
        generator: The generator to debug.
        args: parseargs args object.
    """
    # display images, one at a time
    for i in range(generator.size()):
        # load the data
        image       = generator.load_image(i)
        annotations = generator.load_annotations(i)

        # apply random transformations
        if args.random_transform:
            image, annotations = generator.random_transform_group_entry(image, annotations)

        # resize the image and annotations
        if args.resize:
            image, image_scale = generator.resize_image(image)
            annotations[:, :4] *= image_scale

        # draw anchors on the image
        if args.anchors:
            labels, _, anchors = generator.compute_anchor_targets(image.shape, annotations, generator.num_classes())
            draw_boxes(image, anchors[np.max(labels, axis=1) == 1], (255, 255, 0), thickness=1)

        # draw annotations on the image
        if args.annotations:
            # draw annotations in red
            draw_annotations(image, annotations, color=(0, 0, 255), label_to_name=generator.label_to_name)

            # draw regressed anchors in green to override most red annotations
            # result is that annotations without anchors are red, with anchors are green
            labels, boxes, _ = generator.compute_anchor_targets(image.shape, annotations, generator.num_classes())
            draw_boxes(image, boxes[np.max(labels, axis=1) == 1], (0, 255, 0))

        cv2.imshow('Image', image)
        if cv2.waitKey() == ord('q'):
            return False
    return True


def main(args=None):
    # parse arguments
    if args is None:
        args = sys.argv[1:]
    args = parse_args(args)

    # create the generator
    generator = create_generator(args)

    # create the display window
    cv2.namedWindow('Image', cv2.WINDOW_NORMAL)

    if args.loop:
        while run(generator, args):
            pass
    else:
        run(generator, args)


if __name__ == '__main__':
    main()



class Backbone(object):
    """ This class stores additional information on backbones.
    """
    def __init__(self, backbone):
        # a dictionary mapping custom layer names to the correct classes
        from .. import layers
        from .. import losses
        from .. import initializers
        self.custom_objects = {
            'UpsampleLike'     : layers.UpsampleLike,
            'PriorProbability' : initializers.PriorProbability,
            'RegressBoxes'     : layers.RegressBoxes,
            'FilterDetections' : layers.FilterDetections,
            'Anchors'          : layers.Anchors,
            'ClipBoxes'        : layers.ClipBoxes,
            '_smooth_l1'       : losses.smooth_l1(),
            '_focal'           : losses.focal(),
        }

        self.backbone = backbone
        self.validate()

    def retinanet(self, *args, **kwargs):
        """ Returns a retinanet model using the correct backbone.
        """
        raise NotImplementedError('retinanet method not implemented.')

    def download_imagenet(self):
        """ Downloads ImageNet weights and returns path to weights file.
        """
        raise NotImplementedError('download_imagenet method not implemented.')

    def validate(self):
        """ Checks whether the backbone string is correct.
        """
        raise NotImplementedError('validate method not implemented.')


def backbone(backbone_name):
    """ Returns a backbone object for the given backbone.
    """
    if 'resnet' in backbone_name:
        from .resnet import ResNetBackbone as b
    elif 'mobilenet' in backbone_name:
        from .mobilenet import MobileNetBackbone as b
    elif 'vgg' in backbone_name:
        from .vgg import VGGBackbone as b
    elif 'densenet' in backbone_name:
        from .densenet import DenseNetBackbone as b
    else:
        raise NotImplementedError('Backbone class for  \'{}\' not implemented.'.format(backbone))

    return b(backbone_name)


def load_model(filepath, backbone_name='resnet50', convert=False, nms=True):
    """ Loads a retinanet model using the correct custom objects.

    # Arguments
        filepath: one of the following:
            - string, path to the saved model, or
            - h5py.File object from which to load the model
        backbone_name: Backbone with which the model was trained.
        convert: Boolean, whether to convert the model to an inference model.
        nms: Boolean, whether to add NMS filtering to the converted model. Only valid if convert=True.

    # Returns
        A keras.models.Model object.

    # Raises
        ImportError: if h5py is not available.
        ValueError: In case of an invalid savefile.
    """
    import keras.models

    model = keras.models.load_model(filepath, custom_objects=backbone(backbone_name).custom_objects)
    if convert:
        from .retinanet import retinanet_bbox
        model = retinanet_bbox(model=model, nms=nms)

    return model



"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import keras
import keras_retinanet.layers


class TestAnchors(object):
    def test_simple(self):
        # create simple Anchors layer
        anchors_layer = keras_retinanet.layers.Anchors(
            size=32,
            stride=8,
            ratios=np.array([1], keras.backend.floatx()),
            scales=np.array([1], keras.backend.floatx()),
        )

        # create fake features input (only shape is used anyway)
        features = np.zeros((1, 2, 2, 1024), dtype=keras.backend.floatx())
        features = keras.backend.variable(features)

        # call the Anchors layer
        anchors = anchors_layer.call(features)
        anchors = keras.backend.eval(anchors)

        # expected anchor values
        expected = np.array([[
            [-12, -12, 20, 20],
            [-4 , -12, 28, 20],
            [-12, -4 , 20, 28],
            [-4 , -4 , 28, 28],
        ]], dtype=keras.backend.floatx())

        # test anchor values
        np.testing.assert_array_equal(anchors, expected)

    # mark test to fail
    def test_mini_batch(self):
        # create simple Anchors layer
        anchors_layer = keras_retinanet.layers.Anchors(
            size=32,
            stride=8,
            ratios=np.array([1], dtype=keras.backend.floatx()),
            scales=np.array([1], dtype=keras.backend.floatx()),
        )

        # create fake features input with batch_size=2
        features = np.zeros((2, 2, 2, 1024), dtype=keras.backend.floatx())
        features = keras.backend.variable(features)

        # call the Anchors layer
        anchors = anchors_layer.call(features)
        anchors = keras.backend.eval(anchors)

        # expected anchor values
        expected = np.array([[
            [-12, -12, 20, 20],
            [-4 , -12, 28, 20],
            [-12, -4 , 20, 28],
            [-4 , -4 , 28, 28],
        ]], dtype=keras.backend.floatx())
        expected = np.tile(expected, (2, 1, 1))

        # test anchor values
        np.testing.assert_array_equal(anchors, expected)


class TestUpsampleLike(object):
    def test_simple(self):
        # create simple UpsampleLike layer
        upsample_like_layer = keras_retinanet.layers.UpsampleLike()

        # create input source
        source   = np.zeros((1, 2, 2, 1), dtype=keras.backend.floatx())
        source   = keras.backend.variable(source)
        target   = np.zeros((1, 5, 5, 1), dtype=keras.backend.floatx())
        expected = target
        target   = keras.backend.variable(target)

        # compute output
        actual = upsample_like_layer.call([source, target])
        actual = keras.backend.eval(actual)

        np.testing.assert_array_equal(actual, expected)

    def test_mini_batch(self):
        # create simple UpsampleLike layer
        upsample_like_layer = keras_retinanet.layers.UpsampleLike()

        # create input source
        source = np.zeros((2, 2, 2, 1), dtype=keras.backend.floatx())
        source = keras.backend.variable(source)

        target   = np.zeros((2, 5, 5, 1), dtype=keras.backend.floatx())
        expected = target
        target   = keras.backend.variable(target)

        # compute output
        actual = upsample_like_layer.call([source, target])
        actual = keras.backend.eval(actual)

        np.testing.assert_array_equal(actual, expected)


class TestRegressBoxes(object):
    def test_simple(self):
        mean = [0, 0, 0, 0]
        std  = [0.2, 0.2, 0.2, 0.2]

        # create simple RegressBoxes layer
        regress_boxes_layer = keras_retinanet.layers.RegressBoxes(mean=mean, std=std)

        # create input
        anchors = np.array([[
            [0 , 0 , 10 , 10 ],
            [50, 50, 100, 100],
            [20, 20, 40 , 40 ],
        ]], dtype=keras.backend.floatx())
        anchors = keras.backend.variable(anchors)

        regression = np.array([[
            [0  , 0  , 0  , 0  ],
            [0.1, 0.1, 0  , 0  ],
            [0  , 0  , 0.1, 0.1],
        ]], dtype=keras.backend.floatx())
        regression = keras.backend.variable(regression)

        # compute output
        actual = regress_boxes_layer.call([anchors, regression])
        actual = keras.backend.eval(actual)

        # compute expected output
        expected = np.array([[
            [0 , 0 , 10  , 10  ],
            [51, 51, 100 , 100 ],
            [20, 20, 40.4, 40.4],
        ]], dtype=keras.backend.floatx())

        np.testing.assert_array_almost_equal(actual, expected, decimal=2)

    # mark test to fail
    def test_mini_batch(self):
        mean = [0, 0, 0, 0]
        std  = [0.2, 0.2, 0.2, 0.2]

        # create simple RegressBoxes layer
        regress_boxes_layer = keras_retinanet.layers.RegressBoxes(mean=mean, std=std)

        # create input
        anchors = np.array([
            [
                [0 , 0 , 10 , 10 ],  # 1
                [50, 50, 100, 100],  # 2
                [20, 20, 40 , 40 ],  # 3
            ],
            [
                [20, 20, 40 , 40 ],  # 3
                [0 , 0 , 10 , 10 ],  # 1
                [50, 50, 100, 100],  # 2
            ],
        ], dtype=keras.backend.floatx())
        anchors = keras.backend.variable(anchors)

        regression = np.array([
            [
                [0  , 0  , 0  , 0  ],  # 1
                [0.1, 0.1, 0  , 0  ],  # 2
                [0  , 0  , 0.1, 0.1],  # 3
            ],
            [
                [0  , 0  , 0.1, 0.1],  # 3
                [0  , 0  , 0  , 0  ],  # 1
                [0.1, 0.1, 0  , 0  ],  # 2
            ],
        ], dtype=keras.backend.floatx())
        regression = keras.backend.variable(regression)

        # compute output
        actual = regress_boxes_layer.call([anchors, regression])
        actual = keras.backend.eval(actual)

        # compute expected output
        expected = np.array([
            [
                [0 , 0 , 10  , 10  ],  # 1
                [51, 51, 100 , 100 ],  # 2
                [20, 20, 40.4, 40.4],  # 3
            ],
            [
                [20, 20, 40.4, 40.4],  # 3
                [0 , 0 , 10  , 10  ],  # 1
                [51, 51, 100 , 100 ],  # 2
            ],
        ], dtype=keras.backend.floatx())

        np.testing.assert_array_almost_equal(actual, expected, decimal=2)



"""Python utility which reads entries (different annotations) from a csv file,
    converts them to integer/list of integers(for multilabel-multiclass settings) and saves them in a numpy array.
"""

import pandas
import numpy as np
import math
from keras.utils import np_utils


def csv_to_numpy(csv_path,
                 nb_samples,
                 entry_type,
                 to_file):
    """Reads entries (different annotations) from a csv file,
        converts them to integer/list of integers(for multilabel-multiclass settings) and saves them in a numpy array.

        # Arguments
            csv_path: String to declare the full path of the csv file .
            nb_samples: Number of entries in the csv file to iterate.
            entry_type: The header name of the column to process. One of `emotions`, `valence`, `arousal`, `dominance` and `age`.
            to_file: File name of the numpy array that will hold the converted entries.
        """

    csv_file = pandas.read_csv(csv_path)

    # counter to iterate through all csv entries
    field_number = 0

    final_list = []

    if entry_type == 'emotions':
        for entry in csv_file.emotions:

            # convert the retrieved csv entry (whose type is always str) into integer/list of integers
            int_list = map(int, entry.split(','))

            int_list = np_utils.to_categorical(int_list)

            # append the converted integer/list of integers to the final list
            final_list.append(int_list)


            field_number += 1
            if field_number > nb_samples - 1:
                break

    elif entry_type == 'valence':
        for entry in csv_file.valence:

            # ensure all the nan values for valence, arousal and dominance are transformed to 5.0
            if math.isnan(entry):
                entry = 5.0

            # append the converted integer/list of integers to the final list
            final_list.append(int(entry))

            field_number += 1
            if field_number > nb_samples - 1:
                break

    elif entry_type == 'arousal':
        for entry in csv_file.arousal:

            # ensure all the nan values for valence, arousal and dominance are transformed to 5.0
            if math.isnan(entry):
                entry = 5.0

            final_list.append(int(entry))

            field_number += 1
            if field_number > nb_samples - 1:
                break

    elif entry_type == 'dominance':
        for entry in csv_file.dominance:

            # ensure all the nan values for valence, arousal and dominance are transformed to 5.0
            if math.isnan(entry):
                entry = 5.0

            final_list.append(int(entry))

            field_number += 1
            if field_number > nb_samples - 1:
                break

    elif entry_type == 'age':
        for entry in csv_file.age:

            final_list.append(int(entry))

            field_number += 1
            if field_number > nb_samples - 1:
                break

    # expand dimensions from (xxxx,) to(xxxx, 1)
    final_list = np.expand_dims(final_list, axis=1)
    np.save(to_file, final_list)






if __name__ == '__main__':



    # csv_path= '/home/sandbox/Desktop/EMOTIC_database/RESOURCES/EMOTIC_CSV_FILES/test.csv'
    # nb_samples = 7280
    # entry_type = 'age'
    #
    # to_file= 'age_test'
    #
    # csv_to_numpy(csv_path =csv_path,
    #              nb_samples = nb_samples,
    #              entry_type = entry_type,
    #              to_file= to_file)
    #
    #
    # x = np.load('age_test.npy')
    #
    # # print x.shape
    # #
    # # print x[0]
    #
    # from keras.utils import np_utils
    #
    # x = np_utils.to_categorical(x)
    #
    #
    # np.save(to_file, x)
    #
    # x = np.load('age_test.npy')
    #
    # print x.shape
    # print x[0]
    #
    # print type(x[0][0])


    mode = 'train'
    entry_type = 'age'



    to_file = entry_type + '_'+mode
    csv_path= '/home/sandbox/Desktop/Keras-EMOTIC/dataset/'+ mode +'.csv'


    if mode == 'train':
        nb_samples = 23706
    elif mode == 'val':
        nb_samples = 3334
    elif mode == 'test':
        nb_samples = 7280




    csv_to_numpy(csv_path =csv_path,
                 nb_samples = nb_samples,
                 entry_type = entry_type,
                 to_file= to_file)


    x = np.load(str(to_file)+'.npy')

    print x.shape

    print x[0]

    # from keras.utils import np_utils
    #
    # x = np_utils.to_categorical(x)
    #
    #
    # np.save(to_file, x)
    #
    # x = np.load(str(to_file) + '.npy')
    #
    # print x.shape
    # print x[0]
    #
    # print type(x[0][0])


from __future__ import print_function
from engine.displaced_people_branch.single_image_inference_hra_2class import single_img_HRA_inference


from keras.preprocessing import image
from scipy.misc import imread
from matplotlib import pyplot as plt

from inference.displacenet_single_image_inference_unified import displaceNet_inference


img_path = '/home/sandbox/Desktop/RF234809_Medium_res-LILY-AND-RUEBEN.jpg'
violation_class = 'dp'
hra_model_backend_name = 'VGG16'
nb_of_conv_layers_to_fine_tune = 1

raw_preds, sole_classifier_overlayed_text, top_1_predicted_label = single_img_HRA_inference(img_path=img_path,
                                                                                            violation_class=violation_class,
                                                                                            model_backend_name=hra_model_backend_name,
                                                                                            nb_of_conv_layers_to_fine_tune=nb_of_conv_layers_to_fine_tune)


img = image.load_img(img_path, target_size=(224, 224))
print ('Vanilla CNN prediction: ', raw_preds[0])


emotic_model_a_backend_name = 'VGG16'
emotic_model_b_backend_name = None
emotic_model_c_backend_name = None

final_preds = displaceNet_inference(img_path,
                                    emotic_model_a_backend_name,
                                    emotic_model_b_backend_name,
                                    emotic_model_c_backend_name,
                                    hra_model_backend_name,
                                    nb_of_conv_layers_to_fine_tune,
                                    violation_class)



print('DisplaceNet prediction: ', final_preds)

numpy_img_path = imread(img_path)
plt.figure(figsize=(10, 12))
plt.imshow(numpy_img_path)

current_axis = plt.gca()

# configure colours for bounding box and text
violation_bounding_box_colour_rgbvar = (255, 3, 62)
violation_bounding_box_colour_rgbvar2 = [x / 255.0 for x in violation_bounding_box_colour_rgbvar]

no_violation_bounding_box_colour_rgbvar = (34, 139, 34)
no_violation_bounding_box_colour_rgbvar2 = [x / 255.0 for x in no_violation_bounding_box_colour_rgbvar]

abusenet_overlayed_text = str(final_preds[0][0][1]) + ' (' + str(round(final_preds[0][0][2], 2)) + ')'

# print (abusenet_overlayed_text)

abusenet_overlayed_text = 'DisplaceNet: '+abusenet_overlayed_text
sole_classifier_overlayed_text = 'Vanilla CNN: '+ sole_classifier_overlayed_text


if violation_class == 'dp':
    if final_preds[0][0][1] == 'displaced_populations':
        current_axis.text(0, -28, abusenet_overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        if top_1_predicted_label == 'displaced_populations':
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})
        else:
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

    elif final_preds[0][0][1] == 'no_displaced_populations':
        current_axis.text(0, -45, abusenet_overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        if top_1_predicted_label == 'displaced_populations':
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})
        else:
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

else:
    if final_preds[0][0][1] == 'child_labour':
        current_axis.text(0, -38, abusenet_overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        if top_1_predicted_label == 'child_labour':
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})
        else:
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

    elif final_preds[0][0][1] == 'no_child_labour':
        current_axis.text(0, -38, abusenet_overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        if top_1_predicted_label == 'child_labour':
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})
        else:
            current_axis.text(0, -7, sole_classifier_overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})





plt.axis('off')
plt.show()



# -*- coding: utf-8 -*-
'''

'''

from __future__ import print_function
from engine.human_centric_branch.global_emotional_traits_branch import single_img_VAD_inference_return_only
from engine.displaced_people_branch.single_image_inference_hra_2class import single_img_HRA_inference_return_only
from utils.inference_utils import _obtain_emotional_traits_calibrated_predictions
from applications.hra_utils import decode_predictions

from keras.preprocessing import image
from scipy.misc import imread
from matplotlib import pyplot as plt

def displaceNet_inference(img_path,
                          emotic_model_a_backend_name,
                          emotic_model_b_backend_name,
                          emotic_model_c_backend_name,
                          hra_model_backend_name,
                          nb_of_fine_tuned_conv_layers,
                          violation_class):






    # obtain global emotional traits as VAD values
    global_valence, global_dominance = single_img_VAD_inference_return_only(img_path=img_path,
                                                                            object_detector_backend='RetinaNet',
                                                                            model_a_backend_name=emotic_model_a_backend_name,
                                                                            model_b_backend_name=emotic_model_b_backend_name,
                                                                            model_c_backend_name=emotic_model_c_backend_name,
                                                                            )

    raw_HRA_preds = single_img_HRA_inference_return_only(img_path=img_path,
                                                         violation_class=violation_class,
                                                         model_backend_name=hra_model_backend_name,
                                                         nb_of_conv_layers_to_fine_tune=nb_of_fine_tuned_conv_layers)

    # plain_preds = decode_predictions(violation_class=violation_class,
    #                                  preds=raw_HRA_preds,
    #                                  top=2)

    # # Uncomment for extra verbosity
    # plain_predicted_probability = plain_preds[0][0][2]
    #
    # plain_predicted_label = plain_preds[0][0][1]
    #
    # print ('\n')
    #
    # print('[INFO] Plain predictions: ',plain_predicted_label, '->', plain_predicted_probability)
    #
    # print ('Global dominance: ', global_dominance)

    # Case where no people were detected
    if global_dominance == 0:
        # print ('No people were detected!')


        final_preds = decode_predictions(violation_class=violation_class,
                                         preds=raw_HRA_preds,
                                         top=2)



    else:

        calibrated_preds = _obtain_emotional_traits_calibrated_predictions(emotional_trait=global_dominance,
                                                                           raw_preds=raw_HRA_preds)


        final_preds = decode_predictions(violation_class=violation_class,
                                         preds=calibrated_preds,
                                         top=2)




        # # Uncomment for extra verbosity
        # calibrated_predicted_probability = final_preds[0][0][2]
        #
        # calibrated_predicted_label = final_preds[0][0][1]
        #
        # print('[INFO] Calibrated predictions: ', calibrated_predicted_label, '->', calibrated_predicted_probability)


    return final_preds



if __name__ == "__main__":


    img_path = '/home/sandbox/Desktop/Testing Images/human_right_viol_1.jpg'
    violation_class = 'cl'
    hra_model_backend_name = 'VGG16'
    nb_of_fine_tuned_conv_layers = 1

    emotic_model_a_backend_name = 'VGG16'
    emotic_model_b_backend_name = None
    emotic_model_c_backend_name = None


    final_preds = displaceNet_inference(img_path,
                                        emotic_model_a_backend_name,
                                        emotic_model_b_backend_name,
                                        emotic_model_c_backend_name,
                                        hra_model_backend_name,
                                        nb_of_fine_tuned_conv_layers,
                                        violation_class)

    print (final_preds)

    img = image.load_img(img_path, target_size=(224, 224))

    # plot_preds(violation_class, img, raw_preds[0])


    numpy_img_path = imread(img_path)
    plt.figure(figsize=(10, 12))
    plt.imshow(numpy_img_path)

    current_axis = plt.gca()

    # configure colours for bounding box and text
    violation_bounding_box_colour_rgbvar = (255, 3, 62)
    violation_bounding_box_colour_rgbvar2 = [x / 255.0 for x in violation_bounding_box_colour_rgbvar]

    no_violation_bounding_box_colour_rgbvar = (34, 139, 34)
    no_violation_bounding_box_colour_rgbvar2 = [x / 255.0 for x in no_violation_bounding_box_colour_rgbvar]


    overlayed_text = str(final_preds[0][0][1]) + ' (' + str(round(final_preds[0][0][2], 2)) + ')'

    if violation_class == 'dp':
        if final_preds[0][0][1] == 'displaced_populations':
            current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        elif final_preds[0][0][1] == 'no_displaced_populations':
            current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

    else:
        if final_preds[0][0][1] == 'child_labour':
            current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})

        elif final_preds[0][0][1] == 'no_child_labour':
            current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
                              bbox={'facecolor': no_violation_bounding_box_colour_rgbvar2, 'alpha': 1.0})






    plt.axis('off')
    plt.show()


    # img_path = '/home/sandbox/Desktop/Testing Images/camping.jpg'
    # violation_class = 'dp'
    # hra_model_backend_name = 'VGG16'
    # nb_of_fine_tuned_conv_layers = 1
    #
    # emotic_model_a_backend_name = 'VGG16'
    # emotic_model_b_backend_name = 'VGG19'
    # emotic_model_c_backend_name = 'ResNet50'
    #
    #
    # final_preds = displaceNet_inference(img_path,
    #                    emotic_model_a_backend_name,
    #                    emotic_model_b_backend_name,
    #                    emotic_model_c_backend_name,
    #                    hra_model_backend_name,
    #                    nb_of_fine_tuned_conv_layers,
    #                    violation_class)
    #
    # print (final_preds)


# -*- coding: utf-8 -*-
"""2 clss Human Rights Archive (HRA) VGG19 model for Keras

"""

from __future__ import division, print_function
import os

import warnings
import numpy as np

from keras import backend as K
from keras.utils.data_utils import get_file
from keras.layers import Input
from keras.layers.core import Dense
from keras.models import Model
from keras.layers.core import Dropout
from keras.layers import GlobalAveragePooling2D
from keras.applications.vgg19 import VGG19
from keras.applications.imagenet_utils import _obtain_input_shape
from keras.engine.topology import get_source_inputs
from keras.optimizers import SGD
from applications.hra_utils import _obtain_weights_path as owp

from applications.hra_utils import _obtain_train_mode

pre_trained_model = 'vgg19'

# ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------
CL_WEIGHTS_FEATURE_EXTRACTION_PATH, CL_FEATURE_EXTRACTION_FNAME = owp('cl', pre_trained_model, None, True)
DP_WEIGHTS_FEATURE_EXTRACTION_PATH, DP_FEATURE_EXTRACTION_FNAME = owp('cl', pre_trained_model, None, True)

# ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------
CL_WEIGHTS_PATH_ONE_CONV_LAYER, CL_PATH_ONE_CONV_LAYER_FNAME = owp('cl', pre_trained_model, 1, True)
CL_WEIGHTS_PATH_ONE_CONV_LAYER_NO_TOP, CL_PATH_ONE_CONV_LAYER_NO_TOP_FNAME = owp('cl', pre_trained_model, 1, False)

CL_WEIGHTS_PATH_TWO_CONV_LAYERS, CL_PATH_TWO_CONV_LAYERS_FNAME = owp('cl', pre_trained_model, 2, True)
CL_WEIGHTS_PATH_TWO_CONV_LAYERS_NO_TOP, CL_PATH_TWO_CONV_LAYERS_NO_TOP_FNAME = owp('cl', pre_trained_model, 2, False)

CL_WEIGHTS_PATH_THREE_CONV_LAYERS, CL_PATH_THREE_CONV_LAYERS_FNAME = owp('cl', pre_trained_model, 3, True)
CL_WEIGHTS_PATH_THREE_CONV_LAYERS_NO_TOP, CL_PATH_THREE_CONV_LAYERS_NO_TOP_FNAME = owp('cl', pre_trained_model, 3, False)

# ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------
DP_WEIGHTS_PATH_ONE_CONV_LAYER, DP_PATH_ONE_CONV_LAYER_FNAME = owp('dp', pre_trained_model, 1, True)
DP_WEIGHTS_PATH_ONE_CONV_LAYER_NO_TOP, DP_PATH_ONE_CONV_LAYER_NO_TOP_FNAME = owp('dp', pre_trained_model, 1, False)

DP_WEIGHTS_PATH_TWO_CONV_LAYERS, DP_PATH_TWO_CONV_LAYERS_FNAME = owp('dp', pre_trained_model, 2, True)
DP_WEIGHTS_PATH_TWO_CONV_LAYERS_NO_TOP, DP_PATH_TWO_CONV_LAYERS_NO_TOP_FNAME = owp('dp', pre_trained_model, 2, False)

DP_WEIGHTS_PATH_THREE_CONV_LAYERS, DP_PATH_THREE_CONV_LAYERS_FNAME = owp('dp', pre_trained_model, 3, True)
DP_WEIGHTS_PATH_THREE_CONV_LAYERS_NO_TOP, DP_PATH_THREE_CONV_LAYERS_NO_TOP_FNAME = owp('dp', pre_trained_model, 3, False)

# ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------  ======================  ------------------------

def HRA_VGG19(include_top=True, weights='HRA',
              input_tensor=None, input_shape=None,
              nb_of_conv_layers_to_fine_tune=None,
              first_phase_trained_weights = None,
              violation_class = 'cl',
              verbose=0):
    """Instantiates the VGG19 architecture fine-tuned (2 steps) on Human Rights Archive dataset.

    Optionally loads weights pre-trained on the 2 class version of Human Rights Archive Database.

    # Arguments
        include_top: whether to include the 3 fully-connected
            layers at the top of the network.
        weights: one of `None` (random initialization),
            'HRA' (pre-training on Human Rights Archive),
            or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` data format)
            or `(3, 224, 224)` (with `channels_first` data format).
            It should have exactly 3 input channels,
            and width and height should be no smaller than 48.
            E.g. `(200, 200, 3)` would be one valid value.
        nb_of_conv_layers_to_fine_tune: integer to indicate the number of convolutional
            layers to fine-tune. One of `1` (2,359,808 trainable params), `2` (4,719,616 trainable params) or `3` (7,079,424 trainable params).
        first_phase_trained_weights: Weights of an already trained Keras model instance.
            Only relevant when using `fine_tuning` as train_mode after `feature_extraction` weights have been saved.
        violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
        verbose: Integer. 0, or 1. Verbosity mode. 0 = silent, 1 = model summary and weights info.

    # Returns
        A Keras model instance.

    # Raises
        ValueError: in case of invalid argument for `weights`, `violation_class`, `nb_of_conv_layers_to_fine_tune` or invalid input shape
        """
    if not (weights in {'HRA', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `HRA` '
                         '(pre-training on Human Rights Archive two-class), '
                         'or the path to the weights file to be loaded.')


    if not (violation_class in {'cl', 'dp'}):
        raise ValueError("The `violation_class` argument should be either "
                         "`cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation') "
                         "'or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')")

    if nb_of_conv_layers_to_fine_tune is None and include_top is False:
        raise ValueError('Setting the `include_top` argument as false '
                         'is only relevant when the `nb_of_conv_layers_to_fine_tune` argument is not None (feature extraction), '
                         'otherwise the returned model would be exactly the default '
                         'keras-applications model.')

    if weights == 'HRA' and first_phase_trained_weights is not None:
        raise ValueError('Setting the `first_phase_trained_weights` argument as the path to the weights file '
                         'obtained from utilising feature_extraction '
                         'is only relevant when the `weights` argument is `None`. '
                         'If the `weights` argument is `HRA`, it means the model has already been trained on HRA dataset '
                         'and there is no need to provide a path to the weights file (saved from feature_extraction) to be loaded.')

    if not (nb_of_conv_layers_to_fine_tune in {1, 2, 3, None}):
        raise ValueError('The `nb_of_conv_layers_to_fine_tune` argument should be either '
                         '`None` (indicates feature extraction mode), '
                         '`1`, `2` or `3`. '
                         'More than 3 conv. layers are not supported because the more parameters we are training , '
                         'the more we are at risk of overfitting.')

    cache_subdir = 'AbuseNet'

    mode = _obtain_train_mode(nb_of_conv_layers_to_fine_tune=nb_of_conv_layers_to_fine_tune)

    # Determine proper input shape
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=224,
                                      min_size=48,
                                      data_format=K.image_data_format(),
                                      require_flatten=include_top,
                                      weights=weights)

    if input_tensor is None:
        img_input = Input(shape=input_shape)
    else:
        if not K.is_keras_tensor(input_tensor):
            img_input = Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    # Ensure that the model takes into account any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = get_source_inputs(input_tensor)
    else:
        inputs = img_input


    # create the base pre-trained model
    base_model = VGG19(weights='imagenet', include_top=False, input_tensor=img_input)
    x = base_model.output

    # Classification block - build a classifier model to put on top of the convolutional model
    if include_top:

        # add a global spatial pooling layer (which seems to have the best performance)
        x = GlobalAveragePooling2D(name='GAP')(x)

        # add a fully-connected layer
        x = Dense(256, activation='relu', name='FC1')(x)

        # When random init is enabled, we want to include Dropout,
        # otherwise when loading a pre-trained HRA model we want to omit that layer,
        # so the visualisations are done properly (there is an issue if it is included)
        if weights is None:
            x = Dropout(0.5,name='DROPOUT')(x)
        # and a logistic layer with the number of classes defined by the `classes` argument
        x = Dense(2, activation='softmax', name='PREDICTIONS')(x)

        model = Model(inputs=inputs, outputs=x, name='HRA-2CLASS-VGG19')

    else:
        model = Model(inputs=inputs, outputs=x, name='HRA-2CLASS-VGG19-NO-TOP')
        model.compile(optimizer=SGD(lr=0.0001, momentum=0.9),
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])

        return model




    if mode == 'feature_extraction':

        print('[INFO] Feature extraction mode. \n')

        if verbose == 1:
            print(
                '[INFO] Number of trainable weights before freezing the conv. base of the original pre-trained convnet: '
                '' + str(len(model.trainable_weights)))

        for layer in base_model.layers:
            layer.trainable = False

        if verbose == 1:
            print(
                '[INFO] Number of trainable weights after freezing the conv. base of the original pre-trained convnet: '
                '' + str(len(model.trainable_weights)))

        model.compile(optimizer=SGD(lr=0.0001, momentum=0.9),
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])


    elif mode == 'fine_tuning':


            if nb_of_conv_layers_to_fine_tune == 1:
                # Uncomment for extra verbosity
                # print('[INFO] Fine-tuning of the last one (1) conv. layer. \n')

                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights before unfreezing the last conv. layer of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

                for layer in model.layers[:20]:
                    layer.trainable = False
                for layer in model.layers[20:]:
                    layer.trainable = True

                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights after unfreezing the last conv. layer of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

            elif nb_of_conv_layers_to_fine_tune == 2:
                # Uncomment for extra verbosity
                # print('[INFO] Fine-tuning of the last two (2) conv. layers. \n')
                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights before unfreezing the last two (2) conv. layers of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

                for layer in model.layers[:19]:
                    layer.trainable = False
                for layer in model.layers[19:]:
                    layer.trainable = True

                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights after unfreezing the last two (2) conv. layers of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

            elif nb_of_conv_layers_to_fine_tune == 3:
                # Uncomment for extra verbosity
                # print('[INFO] Fine-tuning of the last three (3) conv. layers. \n')
                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights before unfreezing the last three (3) conv. layers of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

                for layer in model.layers[:18]:
                    layer.trainable = False
                for layer in model.layers[18:]:
                    layer.trainable = True

                if verbose == 1:
                    print(
                        '[INFO] Number of trainable weights after unfreezing the last three (3) conv. layers of the model with the retrained classifier: '
                        '' + str(len(model.trainable_weights)))

            model.compile(optimizer=SGD(lr=0.0001, momentum=0.9),
                          loss='categorical_crossentropy',
                          metrics=['accuracy'])

    if verbose == 1:
        model.summary()

    # load weights
    if weights == 'HRA':

        # Child labour
        if violation_class =='cl':
            if include_top:
                if mode == 'feature_extraction':
                    weights_path = get_file(CL_FEATURE_EXTRACTION_FNAME,
                                            CL_WEIGHTS_FEATURE_EXTRACTION_PATH,
                                            cache_subdir=cache_subdir)

                elif mode == 'fine_tuning':

                    if nb_of_conv_layers_to_fine_tune == 1:
                        weights_path = get_file(CL_PATH_ONE_CONV_LAYER_FNAME,
                                                CL_WEIGHTS_PATH_ONE_CONV_LAYER,
                                                cache_subdir=cache_subdir)
                    elif nb_of_conv_layers_to_fine_tune == 2:
                        weights_path = get_file(CL_PATH_TWO_CONV_LAYERS_FNAME,
                                                CL_WEIGHTS_PATH_TWO_CONV_LAYERS,
                                                cache_subdir=cache_subdir)
                    elif nb_of_conv_layers_to_fine_tune == 3:
                        weights_path = get_file(CL_PATH_THREE_CONV_LAYERS_FNAME,
                                                CL_WEIGHTS_PATH_THREE_CONV_LAYERS,
                                                cache_subdir=cache_subdir)

            # no top
            else:
                if nb_of_conv_layers_to_fine_tune == 1:
                    weights_path = get_file(CL_PATH_ONE_CONV_LAYER_NO_TOP_FNAME,
                                            CL_WEIGHTS_PATH_ONE_CONV_LAYER_NO_TOP,
                                            cache_subdir=cache_subdir)
                elif nb_of_conv_layers_to_fine_tune == 2:
                    weights_path = get_file(CL_PATH_TWO_CONV_LAYERS_NO_TOP_FNAME,
                                            CL_WEIGHTS_PATH_TWO_CONV_LAYERS_NO_TOP,
                                            cache_subdir=cache_subdir)
                elif nb_of_conv_layers_to_fine_tune == 3:
                    weights_path = get_file(CL_PATH_THREE_CONV_LAYERS_NO_TOP_FNAME,
                                            CL_WEIGHTS_PATH_THREE_CONV_LAYERS_NO_TOP,
                                            cache_subdir=cache_subdir)
        # Displaced populations
        elif violation_class == 'dp':
            if include_top:
                if mode == 'feature_extraction':
                    weights_path = get_file(DP_FEATURE_EXTRACTION_FNAME,
                                            DP_WEIGHTS_FEATURE_EXTRACTION_PATH,
                                            cache_subdir=cache_subdir)

                elif mode == 'fine_tuning':

                    if nb_of_conv_layers_to_fine_tune == 1:
                        weights_path = get_file(DP_PATH_ONE_CONV_LAYER_FNAME,
                                                DP_WEIGHTS_PATH_ONE_CONV_LAYER,
                                                cache_subdir=cache_subdir)
                    elif nb_of_conv_layers_to_fine_tune == 2:
                        weights_path = get_file(DP_PATH_TWO_CONV_LAYERS_FNAME,
                                                DP_WEIGHTS_PATH_TWO_CONV_LAYERS,
                                                cache_subdir=cache_subdir)
                    elif nb_of_conv_layers_to_fine_tune == 3:
                        weights_path = get_file(DP_PATH_THREE_CONV_LAYERS_FNAME,
                                                DP_WEIGHTS_PATH_THREE_CONV_LAYERS,
                                                cache_subdir=cache_subdir)

            # no top
            else:
                if nb_of_conv_layers_to_fine_tune == 1:
                    weights_path = get_file(DP_PATH_ONE_CONV_LAYER_NO_TOP_FNAME,
                                            DP_WEIGHTS_PATH_ONE_CONV_LAYER_NO_TOP,
                                            cache_subdir=cache_subdir)
                elif nb_of_conv_layers_to_fine_tune == 2:
                    weights_path = get_file(DP_PATH_TWO_CONV_LAYERS_NO_TOP_FNAME,
                                            DP_WEIGHTS_PATH_TWO_CONV_LAYERS_NO_TOP,
                                            cache_subdir=cache_subdir)
                elif nb_of_conv_layers_to_fine_tune == 3:
                    weights_path = get_file(DP_PATH_THREE_CONV_LAYERS_NO_TOP_FNAME,
                                            DP_WEIGHTS_PATH_THREE_CONV_LAYERS_NO_TOP,
                                            cache_subdir=cache_subdir)

        model.load_weights(weights_path)

    elif weights is not None:
        model.load_weights(weights)

    return model



from __future__ import division, print_function
import os

from keras.layers import Input
from keras.layers.core import Dense
from keras.models import Model
from keras.layers.core import Dropout
from keras.layers import GlobalAveragePooling2D
from keras.utils.data_utils import get_file
from keras import regularizers
from keras.utils import plot_model


from keras.applications.resnet50 import ResNet50
from keras.layers.merge import concatenate
from applications.vgg16_places_365 import VGG16_Places365
from keras.optimizers import SGD

from utils.generic_utils import euclidean_distance_loss, rmse


body_inputs = Input(shape=(224, 224, 3), name='INPUT')
image_inputs = Input(shape=(224, 224, 3), name='INPUT')

# Body module
tmp_model = ResNet50(include_top=False, weights='imagenet', input_tensor=body_inputs, pooling='avg')
# tmp_model.summary()

plot_model(tmp_model, to_file='original_model.png',show_shapes = True)

for i, layer in enumerate(tmp_model.layers):
    print(i, layer.name)

new_model = Model(inputs=tmp_model.input, outputs=tmp_model.get_layer(index=169).output)

for i, layer in enumerate(new_model.layers):
    print(i, layer.name)

# tmp_model.summary()

plot_model(new_model, to_file='intermediate_layer.png',show_shapes = True)







'''
Includes:
* Functions to decode and filter raw SSD model output. These are only needed if the
  SSD model does not have a `DecodeDetections` layer.
* Functions to perform greedy non-maximum suppression

Copyright (C) 2018 Pierluigi Ferrari

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from __future__ import division

import numpy as np

from engine.object_detection_branch.single_shot_detector.bounding_box_utils.bounding_box_utils import iou, convert_coordinates


def greedy_nms(y_pred_decoded, iou_threshold=0.45, coords='corners', border_pixels='half'):
    '''
    Perform greedy non-maximum suppression on the input boxes.

    Greedy NMS works by selecting the box with the highest score and
    removing all boxes around it that are too close to it measured by IoU-similarity.
    Out of the boxes that are left over, once again the one with the highest
    score is selected and so on, until no boxes with too much overlap are left.

    Arguments:
        y_pred_decoded (list): A batch of decoded predictions. For a given batch size `n` this
            is a list of length `n` where each list element is a 2D Numpy array.
            For a batch item with `k` predicted boxes this 2D Numpy array has
            shape `(k, 6)`, where each row contains the coordinates of the respective
            box in the format `[class_id, score, xmin, xmax, ymin, ymax]`.
            Technically, the number of columns doesn't have to be 6, it can be
            arbitrary as long as the first four elements of each row are
            `xmin`, `xmax`, `ymin`, `ymax` (in this order) and the last element
            is the score assigned to the prediction. Note that this function is
            agnostic to the scale of the score or what it represents.
        iou_threshold (float, optional): All boxes with a Jaccard similarity of
            greater than `iou_threshold` with a locally maximal box will be removed
            from the set of predictions, where 'maximal' refers to the box score.
        coords (str, optional): The coordinate format of `y_pred_decoded`.
            Can be one of the formats supported by `iou()`.
        border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
            Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
            to the boxes. If 'exclude', the border pixels do not belong to the boxes.
            If 'half', then one of each of the two horizontal and vertical borders belong
            to the boxex, but not the other.

    Returns:
        The predictions after removing non-maxima. The format is the same as the input format.
    '''
    y_pred_decoded_nms = []
    for batch_item in y_pred_decoded: # For the labels of each batch item...
        boxes_left = np.copy(batch_item)
        maxima = [] # This is where we store the boxes that make it through the non-maximum suppression
        while boxes_left.shape[0] > 0: # While there are still boxes left to compare...
            maximum_index = np.argmax(boxes_left[:,1]) # ...get the index of the next box with the highest confidence...
            maximum_box = np.copy(boxes_left[maximum_index]) # ...copy that box and...
            maxima.append(maximum_box) # ...append it to `maxima` because we'll definitely keep it
            boxes_left = np.delete(boxes_left, maximum_index, axis=0) # Now remove the maximum box from `boxes_left`
            if boxes_left.shape[0] == 0: break # If there are no boxes left after this step, break. Otherwise...
            similarities = iou(boxes_left[:,2:], maximum_box[2:], coords=coords, mode='element-wise', border_pixels=border_pixels) # ...compare (IoU) the other left over boxes to the maximum box...
            boxes_left = boxes_left[similarities <= iou_threshold] # ...so that we can remove the ones that overlap too much with the maximum box
        y_pred_decoded_nms.append(np.array(maxima))

    return y_pred_decoded_nms

def _greedy_nms(predictions, iou_threshold=0.45, coords='corners', border_pixels='half'):
    '''
    The same greedy non-maximum suppression algorithm as above, but slightly modified for use as an internal
    function for per-class NMS in `decode_detections()`.
    '''
    boxes_left = np.copy(predictions)
    maxima = [] # This is where we store the boxes that make it through the non-maximum suppression
    while boxes_left.shape[0] > 0: # While there are still boxes left to compare...
        maximum_index = np.argmax(boxes_left[:,0]) # ...get the index of the next box with the highest confidence...
        maximum_box = np.copy(boxes_left[maximum_index]) # ...copy that box and...
        maxima.append(maximum_box) # ...append it to `maxima` because we'll definitely keep it
        boxes_left = np.delete(boxes_left, maximum_index, axis=0) # Now remove the maximum box from `boxes_left`
        if boxes_left.shape[0] == 0: break # If there are no boxes left after this step, break. Otherwise...
        similarities = iou(boxes_left[:,1:], maximum_box[1:], coords=coords, mode='element-wise', border_pixels=border_pixels) # ...compare (IoU) the other left over boxes to the maximum box...
        boxes_left = boxes_left[similarities <= iou_threshold] # ...so that we can remove the ones that overlap too much with the maximum box
    return np.array(maxima)

def _greedy_nms2(predictions, iou_threshold=0.45, coords='corners', border_pixels='half'):
    '''
    The same greedy non-maximum suppression algorithm as above, but slightly modified for use as an internal
    function in `decode_detections_fast()`.
    '''
    boxes_left = np.copy(predictions)
    maxima = [] # This is where we store the boxes that make it through the non-maximum suppression
    while boxes_left.shape[0] > 0: # While there are still boxes left to compare...
        maximum_index = np.argmax(boxes_left[:,1]) # ...get the index of the next box with the highest confidence...
        maximum_box = np.copy(boxes_left[maximum_index]) # ...copy that box and...
        maxima.append(maximum_box) # ...append it to `maxima` because we'll definitely keep it
        boxes_left = np.delete(boxes_left, maximum_index, axis=0) # Now remove the maximum box from `boxes_left`
        if boxes_left.shape[0] == 0: break # If there are no boxes left after this step, break. Otherwise...
        similarities = iou(boxes_left[:,2:], maximum_box[2:], coords=coords, mode='element-wise', border_pixels=border_pixels) # ...compare (IoU) the other left over boxes to the maximum box...
        boxes_left = boxes_left[similarities <= iou_threshold] # ...so that we can remove the ones that overlap too much with the maximum box
    return np.array(maxima)

def decode_detections(y_pred,
                      confidence_thresh=0.01,
                      iou_threshold=0.45,
                      top_k=200,
                      input_coords='centroids',
                      normalize_coords=True,
                      img_height=None,
                      img_width=None,
                      border_pixels='half'):
    '''
    Convert model prediction output back to a format that contains only the positive box predictions
    (i.e. the same format that `SSDInputEncoder` takes as input).

    After the decoding, two stages of prediction filtering are performed for each class individually:
    First confidence thresholding, then greedy non-maximum suppression. The filtering results for all
    classes are concatenated and the `top_k` overall highest confidence results constitute the final
    predictions for a given batch item. This procedure follows the original Caffe implementation.
    For a slightly different and more efficient alternative to decode raw model output that performs
    non-maximum suppresion globally instead of per class, see `decode_detections_fast()` below.

    Arguments:
        y_pred (array): The prediction output of the SSD model, expected to be a Numpy array
            of shape `(batch_size, #boxes, #classes + 4 + 4 + 4)`, where `#boxes` is the total number of
            boxes predicted by the model per image and the last axis contains
            `[one-hot vector for the classes, 4 predicted coordinate offsets, 4 anchor box coordinates, 4 variances]`.
        confidence_thresh (float, optional): A float in [0,1), the minimum classification confidence in a specific
            positive class in order to be considered for the non-maximum suppression stage for the respective class.
            A lower value will result in a larger part of the selection process being done by the non-maximum suppression
            stage, while a larger value will result in a larger part of the selection process happening in the confidence
            thresholding stage.
        iou_threshold (float, optional): A float in [0,1]. All boxes with a Jaccard similarity of greater than `iou_threshold`
            with a locally maximal box will be removed from the set of predictions for a given class, where 'maximal' refers
            to the box score.
        top_k (int, optional): The number of highest scoring predictions to be kept for each batch item after the
            non-maximum suppression stage.
        input_coords (str, optional): The box coordinate format that the model outputs. Can be either 'centroids'
            for the format `(cx, cy, w, h)` (box center coordinates, width, and height), 'minmax' for the format
            `(xmin, xmax, ymin, ymax)`, or 'corners' for the format `(xmin, ymin, xmax, ymax)`.
        normalize_coords (bool, optional): Set to `True` if the model outputs relative coordinates (i.e. coordinates in [0,1])
            and you wish to transform these relative coordinates back to absolute coordinates. If the model outputs
            relative coordinates, but you do not want to convert them back to absolute coordinates, set this to `False`.
            Do not set this to `True` if the model already outputs absolute coordinates, as that would result in incorrect
            coordinates. Requires `img_height` and `img_width` if set to `True`.
        img_height (int, optional): The height of the input images. Only needed if `normalize_coords` is `True`.
        img_width (int, optional): The width of the input images. Only needed if `normalize_coords` is `True`.
        border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
            Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
            to the boxes. If 'exclude', the border pixels do not belong to the boxes.
            If 'half', then one of each of the two horizontal and vertical borders belong
            to the boxex, but not the other.

    Returns:
        A python list of length `batch_size` where each list element represents the predicted boxes
        for one image and contains a Numpy array of shape `(boxes, 6)` where each row is a box prediction for
        a non-background class for the respective image in the format `[class_id, confidence, xmin, ymin, xmax, ymax]`.
    '''
    if normalize_coords and ((img_height is None) or (img_width is None)):
        raise ValueError("If relative box coordinates are supposed to be converted to absolute coordinates, the decoder needs the image size in order to decode the predictions, but `img_height == {}` and `img_width == {}`".format(img_height, img_width))

    # 1: Convert the box coordinates from the predicted anchor box offsets to predicted absolute coordinates

    y_pred_decoded_raw = np.copy(y_pred[:,:,:-8]) # Slice out the classes and the four offsets, throw away the anchor coordinates and variances, resulting in a tensor of shape `[batch, n_boxes, n_classes + 4 coordinates]`

    if input_coords == 'centroids':
        y_pred_decoded_raw[:,:,[-2,-1]] = np.exp(y_pred_decoded_raw[:,:,[-2,-1]] * y_pred[:,:,[-2,-1]]) # exp(ln(w(pred)/w(anchor)) / w_variance * w_variance) == w(pred) / w(anchor), exp(ln(h(pred)/h(anchor)) / h_variance * h_variance) == h(pred) / h(anchor)
        y_pred_decoded_raw[:,:,[-2,-1]] *= y_pred[:,:,[-6,-5]] # (w(pred) / w(anchor)) * w(anchor) == w(pred), (h(pred) / h(anchor)) * h(anchor) == h(pred)
        y_pred_decoded_raw[:,:,[-4,-3]] *= y_pred[:,:,[-4,-3]] * y_pred[:,:,[-6,-5]] # (delta_cx(pred) / w(anchor) / cx_variance) * cx_variance * w(anchor) == delta_cx(pred), (delta_cy(pred) / h(anchor) / cy_variance) * cy_variance * h(anchor) == delta_cy(pred)
        y_pred_decoded_raw[:,:,[-4,-3]] += y_pred[:,:,[-8,-7]] # delta_cx(pred) + cx(anchor) == cx(pred), delta_cy(pred) + cy(anchor) == cy(pred)
        y_pred_decoded_raw = convert_coordinates(y_pred_decoded_raw, start_index=-4, conversion='centroids2corners')
    elif input_coords == 'minmax':
        y_pred_decoded_raw[:,:,-4:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_decoded_raw[:,:,[-4,-3]] *= np.expand_dims(y_pred[:,:,-7] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_decoded_raw[:,:,[-2,-1]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-6], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_decoded_raw[:,:,-4:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
        y_pred_decoded_raw = convert_coordinates(y_pred_decoded_raw, start_index=-4, conversion='minmax2corners')
    elif input_coords == 'corners':
        y_pred_decoded_raw[:,:,-4:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_decoded_raw[:,:,[-4,-2]] *= np.expand_dims(y_pred[:,:,-6] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_decoded_raw[:,:,[-3,-1]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-7], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_decoded_raw[:,:,-4:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
    else:
        raise ValueError("Unexpected value for `input_coords`. Supported input coordinate formats are 'minmax', 'corners' and 'centroids'.")

    # 2: If the model predicts normalized box coordinates and they are supposed to be converted back to absolute coordinates, do that

    if normalize_coords:
        y_pred_decoded_raw[:,:,[-4,-2]] *= img_width # Convert xmin, xmax back to absolute coordinates
        y_pred_decoded_raw[:,:,[-3,-1]] *= img_height # Convert ymin, ymax back to absolute coordinates

    # 3: Apply confidence thresholding and non-maximum suppression per class

    n_classes = y_pred_decoded_raw.shape[-1] - 4 # The number of classes is the length of the last axis minus the four box coordinates

    y_pred_decoded = [] # Store the final predictions in this list
    for batch_item in y_pred_decoded_raw: # `batch_item` has shape `[n_boxes, n_classes + 4 coords]`
        pred = [] # Store the final predictions for this batch item here
        for class_id in range(1, n_classes): # For each class except the background class (which has class ID 0)...
            single_class = batch_item[:,[class_id, -4, -3, -2, -1]] # ...keep only the confidences for that class, making this an array of shape `[n_boxes, 5]` and...
            threshold_met = single_class[single_class[:,0] > confidence_thresh] # ...keep only those boxes with a confidence above the set threshold.
            if threshold_met.shape[0] > 0: # If any boxes made the threshold...
                maxima = _greedy_nms(threshold_met, iou_threshold=iou_threshold, coords='corners', border_pixels=border_pixels) # ...perform NMS on them.
                maxima_output = np.zeros((maxima.shape[0], maxima.shape[1] + 1)) # Expand the last dimension by one element to have room for the class ID. This is now an arrray of shape `[n_boxes, 6]`
                maxima_output[:,0] = class_id # Write the class ID to the first column...
                maxima_output[:,1:] = maxima # ...and write the maxima to the other columns...
                pred.append(maxima_output) # ...and append the maxima for this class to the list of maxima for this batch item.
        # Once we're through with all classes, keep only the `top_k` maxima with the highest scores
        if pred: # If there are any predictions left after confidence-thresholding...
            pred = np.concatenate(pred, axis=0)
            if top_k != 'all' and pred.shape[0] > top_k: # If we have more than `top_k` results left at this point, otherwise there is nothing to filter,...
                top_k_indices = np.argpartition(pred[:,1], kth=pred.shape[0]-top_k, axis=0)[pred.shape[0]-top_k:] # ...get the indices of the `top_k` highest-score maxima...
                pred = pred[top_k_indices] # ...and keep only those entries of `pred`...
        else:
            pred = np.array(pred) # Even if empty, `pred` must become a Numpy array.
        y_pred_decoded.append(pred) # ...and now that we're done, append the array of final predictions for this batch item to the output list

    return y_pred_decoded

def decode_detections_fast(y_pred,
                           confidence_thresh=0.5,
                           iou_threshold=0.45,
                           top_k='all',
                           input_coords='centroids',
                           normalize_coords=True,
                           img_height=None,
                           img_width=None,
                           border_pixels='half'):
    '''
    Convert model prediction output back to a format that contains only the positive box predictions
    (i.e. the same format that `enconde_y()` takes as input).

    Optionally performs confidence thresholding and greedy non-maximum suppression after the decoding stage.

    Note that the decoding procedure used here is not the same as the procedure used in the original Caffe implementation.
    For each box, the procedure used here assigns the box's highest confidence as its predicted class. Then it removes
    all boxes for which the highest confidence is the background class. This results in less work for the subsequent
    non-maximum suppression, because the vast majority of the predictions will be filtered out just by the fact that
    their highest confidence is for the background class. It is much more efficient than the procedure of the original
    implementation, but the results may also differ.

    Arguments:
        y_pred (array): The prediction output of the SSD model, expected to be a Numpy array
            of shape `(batch_size, #boxes, #classes + 4 + 4 + 4)`, where `#boxes` is the total number of
            boxes predicted by the model per image and the last axis contains
            `[one-hot vector for the classes, 4 predicted coordinate offsets, 4 anchor box coordinates, 4 variances]`.
        confidence_thresh (float, optional): A float in [0,1), the minimum classification confidence in any positive
            class required for a given box to be considered a positive prediction. A lower value will result
            in better recall, while a higher value will result in better precision. Do not use this parameter with the
            goal to combat the inevitably many duplicates that an SSD will produce, the subsequent non-maximum suppression
            stage will take care of those.
        iou_threshold (float, optional): `None` or a float in [0,1]. If `None`, no non-maximum suppression will be
            performed. If not `None`, greedy NMS will be performed after the confidence thresholding stage, meaning
            all boxes with a Jaccard similarity of greater than `iou_threshold` with a locally maximal box will be removed
            from the set of predictions, where 'maximal' refers to the box score.
        top_k (int, optional): 'all' or an integer with number of highest scoring predictions to be kept for each batch item
            after the non-maximum suppression stage. If 'all', all predictions left after the NMS stage will be kept.
        input_coords (str, optional): The box coordinate format that the model outputs. Can be either 'centroids'
            for the format `(cx, cy, w, h)` (box center coordinates, width, and height), 'minmax' for the format
            `(xmin, xmax, ymin, ymax)`, or 'corners' for the format `(xmin, ymin, xmax, ymax)`.
        normalize_coords (bool, optional): Set to `True` if the model outputs relative coordinates (i.e. coordinates in [0,1])
            and you wish to transform these relative coordinates back to absolute coordinates. If the model outputs
            relative coordinates, but you do not want to convert them back to absolute coordinates, set this to `False`.
            Do not set this to `True` if the model already outputs absolute coordinates, as that would result in incorrect
            coordinates. Requires `img_height` and `img_width` if set to `True`.
        img_height (int, optional): The height of the input images. Only needed if `normalize_coords` is `True`.
        img_width (int, optional): The width of the input images. Only needed if `normalize_coords` is `True`.
        border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
            Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
            to the boxes. If 'exclude', the border pixels do not belong to the boxes.
            If 'half', then one of each of the two horizontal and vertical borders belong
            to the boxex, but not the other.

    Returns:
        A python list of length `batch_size` where each list element represents the predicted boxes
        for one image and contains a Numpy array of shape `(boxes, 6)` where each row is a box prediction for
        a non-background class for the respective image in the format `[class_id, confidence, xmin, xmax, ymin, ymax]`.
    '''
    if normalize_coords and ((img_height is None) or (img_width is None)):
        raise ValueError("If relative box coordinates are supposed to be converted to absolute coordinates, the decoder needs the image size in order to decode the predictions, but `img_height == {}` and `img_width == {}`".format(img_height, img_width))

    # 1: Convert the classes from one-hot encoding to their class ID
    y_pred_converted = np.copy(y_pred[:,:,-14:-8]) # Slice out the four offset predictions plus two elements whereto we'll write the class IDs and confidences in the next step
    y_pred_converted[:,:,0] = np.argmax(y_pred[:,:,:-12], axis=-1) # The indices of the highest confidence values in the one-hot class vectors are the class ID
    y_pred_converted[:,:,1] = np.amax(y_pred[:,:,:-12], axis=-1) # Store the confidence values themselves, too

    # 2: Convert the box coordinates from the predicted anchor box offsets to predicted absolute coordinates
    if input_coords == 'centroids':
        y_pred_converted[:,:,[4,5]] = np.exp(y_pred_converted[:,:,[4,5]] * y_pred[:,:,[-2,-1]]) # exp(ln(w(pred)/w(anchor)) / w_variance * w_variance) == w(pred) / w(anchor), exp(ln(h(pred)/h(anchor)) / h_variance * h_variance) == h(pred) / h(anchor)
        y_pred_converted[:,:,[4,5]] *= y_pred[:,:,[-6,-5]] # (w(pred) / w(anchor)) * w(anchor) == w(pred), (h(pred) / h(anchor)) * h(anchor) == h(pred)
        y_pred_converted[:,:,[2,3]] *= y_pred[:,:,[-4,-3]] * y_pred[:,:,[-6,-5]] # (delta_cx(pred) / w(anchor) / cx_variance) * cx_variance * w(anchor) == delta_cx(pred), (delta_cy(pred) / h(anchor) / cy_variance) * cy_variance * h(anchor) == delta_cy(pred)
        y_pred_converted[:,:,[2,3]] += y_pred[:,:,[-8,-7]] # delta_cx(pred) + cx(anchor) == cx(pred), delta_cy(pred) + cy(anchor) == cy(pred)
        y_pred_converted = convert_coordinates(y_pred_converted, start_index=-4, conversion='centroids2corners')
    elif input_coords == 'minmax':
        y_pred_converted[:,:,2:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_converted[:,:,[2,3]] *= np.expand_dims(y_pred[:,:,-7] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_converted[:,:,[4,5]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-6], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_converted[:,:,2:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
        y_pred_converted = convert_coordinates(y_pred_converted, start_index=-4, conversion='minmax2corners')
    elif input_coords == 'corners':
        y_pred_converted[:,:,2:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_converted[:,:,[2,4]] *= np.expand_dims(y_pred[:,:,-6] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_converted[:,:,[3,5]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-7], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_converted[:,:,2:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
    else:
        raise ValueError("Unexpected value for `coords`. Supported values are 'minmax', 'corners' and 'centroids'.")

    # 3: If the model predicts normalized box coordinates and they are supposed to be converted back to absolute coordinates, do that
    if normalize_coords:
        y_pred_converted[:,:,[2,4]] *= img_width # Convert xmin, xmax back to absolute coordinates
        y_pred_converted[:,:,[3,5]] *= img_height # Convert ymin, ymax back to absolute coordinates

    # 4: Decode our huge `(batch, #boxes, 6)` tensor into a list of length `batch` where each list entry is an array containing only the positive predictions
    y_pred_decoded = []
    for batch_item in y_pred_converted: # For each image in the batch...
        boxes = batch_item[np.nonzero(batch_item[:,0])] # ...get all boxes that don't belong to the background class,...
        boxes = boxes[boxes[:,1] >= confidence_thresh] # ...then filter out those positive boxes for which the prediction confidence is too low and after that...
        if iou_threshold: # ...if an IoU threshold is set...
            boxes = _greedy_nms2(boxes, iou_threshold=iou_threshold, coords='corners', border_pixels=border_pixels) # ...perform NMS on the remaining boxes.
        if top_k != 'all' and boxes.shape[0] > top_k: # If we have more than `top_k` results left at this point...
            top_k_indices = np.argpartition(boxes[:,1], kth=boxes.shape[0]-top_k, axis=0)[boxes.shape[0]-top_k:] # ...get the indices of the `top_k` highest-scoring boxes...
            boxes = boxes[top_k_indices] # ...and keep only those boxes...
        y_pred_decoded.append(boxes) # ...and now that we're done, append the array of final predictions for this batch item to the output list

    return y_pred_decoded

################################################################################################
# Debugging tools, not relevant for normal use
################################################################################################

# The functions below are for debugging, so you won't normally need them. That is,
# unless you need to debug your model, of course.

def decode_detections_debug(y_pred,
                            confidence_thresh=0.01,
                            iou_threshold=0.45,
                            top_k=200,
                            input_coords='centroids',
                            normalize_coords=True,
                            img_height=None,
                            img_width=None,
                            variance_encoded_in_target=False,
                            border_pixels='half'):
    '''
    This decoder performs the same processing as `decode_detections()`, but the output format for each left-over
    predicted box is `[box_id, class_id, confidence, xmin, ymin, xmax, ymax]`.

    That is, in addition to the usual data, each predicted box has the internal index of that box within
    the model (`box_id`) prepended to it. This allows you to know exactly which part of the model made a given
    box prediction; in particular, it allows you to know which predictor layer made a given prediction.
    This can be useful for debugging.

    Arguments:
        y_pred (array): The prediction output of the SSD model, expected to be a Numpy array
            of shape `(batch_size, #boxes, #classes + 4 + 4 + 4)`, where `#boxes` is the total number of
            boxes predicted by the model per image and the last axis contains
            `[one-hot vector for the classes, 4 predicted coordinate offsets, 4 anchor box coordinates, 4 variances]`.
        confidence_thresh (float, optional): A float in [0,1), the minimum classification confidence in a specific
            positive class in order to be considered for the non-maximum suppression stage for the respective class.
            A lower value will result in a larger part of the selection process being done by the non-maximum suppression
            stage, while a larger value will result in a larger part of the selection process happening in the confidence
            thresholding stage.
        iou_threshold (float, optional): A float in [0,1]. All boxes with a Jaccard similarity of greater than `iou_threshold`
            with a locally maximal box will be removed from the set of predictions for a given class, where 'maximal' refers
            to the box score.
        top_k (int, optional): The number of highest scoring predictions to be kept for each batch item after the
            non-maximum suppression stage.
        input_coords (str, optional): The box coordinate format that the model outputs. Can be either 'centroids'
            for the format `(cx, cy, w, h)` (box center coordinates, width, and height), 'minmax' for the format
            `(xmin, xmax, ymin, ymax)`, or 'corners' for the format `(xmin, ymin, xmax, ymax)`.
        normalize_coords (bool, optional): Set to `True` if the model outputs relative coordinates (i.e. coordinates in [0,1])
            and you wish to transform these relative coordinates back to absolute coordinates. If the model outputs
            relative coordinates, but you do not want to convert them back to absolute coordinates, set this to `False`.
            Do not set this to `True` if the model already outputs absolute coordinates, as that would result in incorrect
            coordinates. Requires `img_height` and `img_width` if set to `True`.
        img_height (int, optional): The height of the input images. Only needed if `normalize_coords` is `True`.
        img_width (int, optional): The width of the input images. Only needed if `normalize_coords` is `True`.
        border_pixels (str, optional): How to treat the border pixels of the bounding boxes.
            Can be 'include', 'exclude', or 'half'. If 'include', the border pixels belong
            to the boxes. If 'exclude', the border pixels do not belong to the boxes.
            If 'half', then one of each of the two horizontal and vertical borders belong
            to the boxex, but not the other.

    Returns:
        A python list of length `batch_size` where each list element represents the predicted boxes
        for one image and contains a Numpy array of shape `(boxes, 7)` where each row is a box prediction for
        a non-background class for the respective image in the format `[box_id, class_id, confidence, xmin, ymin, xmax, ymax]`.
    '''
    if normalize_coords and ((img_height is None) or (img_width is None)):
        raise ValueError("If relative box coordinates are supposed to be converted to absolute coordinates, the decoder needs the image size in order to decode the predictions, but `img_height == {}` and `img_width == {}`".format(img_height, img_width))

    # 1: Convert the box coordinates from the predicted anchor box offsets to predicted absolute coordinates

    y_pred_decoded_raw = np.copy(y_pred[:,:,:-8]) # Slice out the classes and the four offsets, throw away the anchor coordinates and variances, resulting in a tensor of shape `[batch, n_boxes, n_classes + 4 coordinates]`

    if input_coords == 'centroids':
        if variance_encoded_in_target:
            # Decode the predicted box center x and y coordinates.
            y_pred_decoded_raw[:,:,[-4,-3]] = y_pred_decoded_raw[:,:,[-4,-3]] * y_pred[:,:,[-6,-5]] + y_pred[:,:,[-8,-7]]
            # Decode the predicted box width and heigt.
            y_pred_decoded_raw[:,:,[-2,-1]] = np.exp(y_pred_decoded_raw[:,:,[-2,-1]]) * y_pred[:,:,[-6,-5]]
        else:
            # Decode the predicted box center x and y coordinates.
            y_pred_decoded_raw[:,:,[-4,-3]] = y_pred_decoded_raw[:,:,[-4,-3]] * y_pred[:,:,[-6,-5]] * y_pred[:,:,[-4,-3]] + y_pred[:,:,[-8,-7]]
            # Decode the predicted box width and heigt.
            y_pred_decoded_raw[:,:,[-2,-1]] = np.exp(y_pred_decoded_raw[:,:,[-2,-1]] * y_pred[:,:,[-2,-1]]) * y_pred[:,:,[-6,-5]]
        y_pred_decoded_raw = convert_coordinates(y_pred_decoded_raw, start_index=-4, conversion='centroids2corners')
    elif input_coords == 'minmax':
        y_pred_decoded_raw[:,:,-4:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_decoded_raw[:,:,[-4,-3]] *= np.expand_dims(y_pred[:,:,-7] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_decoded_raw[:,:,[-2,-1]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-6], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_decoded_raw[:,:,-4:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
        y_pred_decoded_raw = convert_coordinates(y_pred_decoded_raw, start_index=-4, conversion='minmax2corners')
    elif input_coords == 'corners':
        y_pred_decoded_raw[:,:,-4:] *= y_pred[:,:,-4:] # delta(pred) / size(anchor) / variance * variance == delta(pred) / size(anchor) for all four coordinates, where 'size' refers to w or h, respectively
        y_pred_decoded_raw[:,:,[-4,-2]] *= np.expand_dims(y_pred[:,:,-6] - y_pred[:,:,-8], axis=-1) # delta_xmin(pred) / w(anchor) * w(anchor) == delta_xmin(pred), delta_xmax(pred) / w(anchor) * w(anchor) == delta_xmax(pred)
        y_pred_decoded_raw[:,:,[-3,-1]] *= np.expand_dims(y_pred[:,:,-5] - y_pred[:,:,-7], axis=-1) # delta_ymin(pred) / h(anchor) * h(anchor) == delta_ymin(pred), delta_ymax(pred) / h(anchor) * h(anchor) == delta_ymax(pred)
        y_pred_decoded_raw[:,:,-4:] += y_pred[:,:,-8:-4] # delta(pred) + anchor == pred for all four coordinates
    else:
        raise ValueError("Unexpected value for `input_coords`. Supported input coordinate formats are 'minmax', 'corners' and 'centroids'.")

    # 2: If the model predicts normalized box coordinates and they are supposed to be converted back to absolute coordinates, do that

    if normalize_coords:
        y_pred_decoded_raw[:,:,[-4,-2]] *= img_width # Convert xmin, xmax back to absolute coordinates
        y_pred_decoded_raw[:,:,[-3,-1]] *= img_height # Convert ymin, ymax back to absolute coordinates

    # 3: For each batch item, prepend each box's internal index to its coordinates.

    y_pred_decoded_raw2 = np.zeros((y_pred_decoded_raw.shape[0], y_pred_decoded_raw.shape[1], y_pred_decoded_raw.shape[2] + 1)) # Expand the last axis by one.
    y_pred_decoded_raw2[:,:,1:] = y_pred_decoded_raw
    y_pred_decoded_raw2[:,:,0] = np.arange(y_pred_decoded_raw.shape[1]) # Put the box indices as the first element for each box via broadcasting.
    y_pred_decoded_raw = y_pred_decoded_raw2

    # 4: Apply confidence thresholding and non-maximum suppression per class

    n_classes = y_pred_decoded_raw.shape[-1] - 5 # The number of classes is the length of the last axis minus the four box coordinates and minus the index

    y_pred_decoded = [] # Store the final predictions in this list
    for batch_item in y_pred_decoded_raw: # `batch_item` has shape `[n_boxes, n_classes + 4 coords]`
        pred = [] # Store the final predictions for this batch item here
        for class_id in range(1, n_classes): # For each class except the background class (which has class ID 0)...
            single_class = batch_item[:,[0, class_id + 1, -4, -3, -2, -1]] # ...keep only the confidences for that class, making this an array of shape `[n_boxes, 6]` and...
            threshold_met = single_class[single_class[:,1] > confidence_thresh] # ...keep only those boxes with a confidence above the set threshold.
            if threshold_met.shape[0] > 0: # If any boxes made the threshold...
                maxima = _greedy_nms_debug(threshold_met, iou_threshold=iou_threshold, coords='corners', border_pixels=border_pixels) # ...perform NMS on them.
                maxima_output = np.zeros((maxima.shape[0], maxima.shape[1] + 1)) # Expand the last dimension by one element to have room for the class ID. This is now an arrray of shape `[n_boxes, 6]`
                maxima_output[:,0] = maxima[:,0] # Write the box index to the first column...
                maxima_output[:,1] = class_id # ...and write the class ID to the second column...
                maxima_output[:,2:] = maxima[:,1:] # ...and write the rest of the maxima data to the other columns...
                pred.append(maxima_output) # ...and append the maxima for this class to the list of maxima for this batch item.
        # Once we're through with all classes, keep only the `top_k` maxima with the highest scores
        pred = np.concatenate(pred, axis=0)
        if pred.shape[0] > top_k: # If we have more than `top_k` results left at this point, otherwise there is nothing to filter,...
            top_k_indices = np.argpartition(pred[:,2], kth=pred.shape[0]-top_k, axis=0)[pred.shape[0]-top_k:] # ...get the indices of the `top_k` highest-score maxima...
            pred = pred[top_k_indices] # ...and keep only those entries of `pred`...
        y_pred_decoded.append(pred) # ...and now that we're done, append the array of final predictions for this batch item to the output list

    return y_pred_decoded

def _greedy_nms_debug(predictions, iou_threshold=0.45, coords='corners', border_pixels='half'):
    '''
    The same greedy non-maximum suppression algorithm as above, but slightly modified for use as an internal
    function for per-class NMS in `decode_detections_debug()`. The difference is that it keeps the indices of all
    left-over boxes for each batch item, which allows you to know which predictor layer predicted a given output
    box and is thus useful for debugging.
    '''
    boxes_left = np.copy(predictions)
    maxima = [] # This is where we store the boxes that make it through the non-maximum suppression
    while boxes_left.shape[0] > 0: # While there are still boxes left to compare...
        maximum_index = np.argmax(boxes_left[:,1]) # ...get the index of the next box with the highest confidence...
        maximum_box = np.copy(boxes_left[maximum_index]) # ...copy that box and...
        maxima.append(maximum_box) # ...append it to `maxima` because we'll definitely keep it
        boxes_left = np.delete(boxes_left, maximum_index, axis=0) # Now remove the maximum box from `boxes_left`
        if boxes_left.shape[0] == 0: break # If there are no boxes left after this step, break. Otherwise...
        similarities = iou(boxes_left[:,2:], maximum_box[2:], coords=coords, mode='element-wise', border_pixels=border_pixels) # ...compare (IoU) the other left over boxes to the maximum box...
        boxes_left = boxes_left[similarities <= iou_threshold] # ...so that we can remove the ones that overlap too much with the maximum box
    return np.array(maxima)

def get_num_boxes_per_pred_layer(predictor_sizes, aspect_ratios, two_boxes_for_ar1):
    '''
    Returns a list of the number of boxes that each predictor layer predicts.

    `aspect_ratios` must be a nested list, containing a list of aspect ratios
    for each predictor layer.
    '''
    num_boxes_per_pred_layer = []
    for i in range(len(predictor_sizes)):
        if two_boxes_for_ar1:
            num_boxes_per_pred_layer.append(predictor_sizes[i][0] * predictor_sizes[i][1] * (len(aspect_ratios[i]) + 1))
        else:
            num_boxes_per_pred_layer.append(predictor_sizes[i][0] * predictor_sizes[i][1] * len(aspect_ratios[i]))
    return num_boxes_per_pred_layer

def get_pred_layers(y_pred_decoded, num_boxes_per_pred_layer):
    '''
    For a given prediction tensor decoded with `decode_detections_debug()`, returns a list
    with the indices of the predictor layers that made each predictions.

    That is, this function lets you know which predictor layer is responsible
    for a given prediction.

    Arguments:
        y_pred_decoded (array): The decoded model output tensor. Must have been
            decoded with `decode_detections_debug()` so that it contains the internal box index
            for each predicted box.
        num_boxes_per_pred_layer (list): A list that contains the total number
            of boxes that each predictor layer predicts.
    '''
    pred_layers_all = []
    cum_boxes_per_pred_layer = np.cumsum(num_boxes_per_pred_layer)
    for batch_item in y_pred_decoded:
        pred_layers = []
        for prediction in batch_item:
            if (prediction[0] < 0) or (prediction[0] >= cum_boxes_per_pred_layer[-1]):
                raise ValueError("Box index is out of bounds of the possible indices as given by the values in `num_boxes_per_pred_layer`.")
            for i in range(len(cum_boxes_per_pred_layer)):
                if prediction[0] < cum_boxes_per_pred_layer[i]:
                    pred_layers.append(i)
                    break
        pred_layers_all.append(pred_layers)
    return pred_layers_all
